"""
A module containing various channel models.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike, verify_condition, verify_scalar
from .._link_budget._capacity import Hb


@export
def bsc(
    x: npt.ArrayLike,
    p: float,
    seed: int | None = None,
) -> npt.NDArray[np.int_]:
    r"""
    Passes the binary input sequence $x$ through a binary symmetric channel (BSC).

    Arguments:
        x: The input sequence $x$ with $x_i \in \{0, 1\}$.
        p: The probability $p$ of a bit flip.
        seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.

    Returns:
        The output sequence $y$ with $y_i \in \{0, 1\}$.

    Examples:
        When 20 bits are passed through a BSC with transition probability $p=0.25$,
        roughly 5 bits are flipped at the output.

        .. ipython:: python

            x = np.random.randint(0, 2, 20); x
            y = sdr.bsc(x, 0.25); y
            x == y

    Group:
        simulation-channel-models
    """
    x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=2)
    verify_scalar(p, float=True, inclusive_min=0, inclusive_max=1)

    rng = np.random.default_rng(seed)
    flip = rng.choice([0, 1], size=x.shape, p=[1 - p, p])
    y = x ^ flip

    return convert_output(y)


@export
def bec(
    x: npt.ArrayLike,
    p: float,
    seed: int | None = None,
) -> npt.NDArray[np.int_]:
    r"""
    Passes the binary input sequence $x$ through a binary erasure channel (BEC).

    Arguments:
        x: The input sequence $x$ with $x_i \in \{0, 1\}$.
        p: The probability $p$ of a bit erasure.
        seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.

    Returns:
        The output sequence $y$ with $y_i \in \{0, 1, e\}$. Erasures $e$ are represented by -1.

    Examples:
        When 20 bits are passed through a BEC with erasure probability $p=0.25$,
        roughly 5 bits are erased at the output.

        .. ipython:: python

            x = np.random.randint(0, 2, 20); x
            y = sdr.bec(x, 0.25); y
            x == y

    Group:
        simulation-channel-models
    """
    x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=2)
    verify_scalar(p, float=True, inclusive_min=0, inclusive_max=1)

    rng = np.random.default_rng(seed)
    random_p = rng.random(x.shape)
    y = np.where(random_p < p, -1, x)

    return convert_output(y)


@export
def dmc(
    x: npt.ArrayLike,
    P: npt.ArrayLike,
    X: npt.ArrayLike | None = None,
    Y: npt.ArrayLike | None = None,
    seed: int | None = None,
) -> npt.NDArray[np.int_]:
    r"""
    Passes the input sequence $x$ through a discrete memoryless channel (DMC).

    Arguments:
        x: The input sequence $x$ with $x_i \in \mathcal{X}$.
        P: The $m \times n$ transition probability matrix $P$, where $P_{i,j} = \Pr(Y = y_j | X = x_i)$.
        X: The input alphabet $\mathcal{X}$ of size $m$. If `None`, it is assumed that
            $\mathcal{X} = \{0, 1, \ldots, m-1\}$.
        Y: The output alphabet $\mathcal{Y}$ of size $n$. If `None`, it is assumed that
            $\mathcal{Y} = \{0, 1, \ldots, n-1\}$.
        seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.

    Returns:
        The output sequence $y$ with $y_i \in \mathcal{Y}$.

    Examples:
        Define the binary symmetric channel (BSC) with transition probability $p=0.25$.

        .. ipython:: python

            p = 0.25
            x = np.random.randint(0, 2, 20); x
            P = [[1 - p, p], [p, 1 - p]]
            y = sdr.dmc(x, P); y
            x == y

        Define the binary erasure channel (BEC) with erasure probability $p=0.25$.

        .. ipython:: python

            p = 0.25
            x = np.random.randint(0, 2, 20); x
            P = [[1 - p, 0, p], [0, 1 - p, p]]
            # Specify the erasure as -1
            y = sdr.dmc(x, P, Y=[0, 1, -1]); y
            x == y

    Group:
        simulation-channel-models
    """
    x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=2)
    P = verify_arraylike(P, float=True, ndim=2, inclusive_min=0, inclusive_max=1)
    verify_condition(np.allclose(P.sum(axis=1), 1))

    if X is None:
        X = np.arange(P.shape[0])
    else:
        X = verify_arraylike(X, int=True, ndim=1, size=P.shape[0])

    if Y is None:
        Y = np.arange(P.shape[1])
    else:
        Y = verify_arraylike(Y, int=True, ndim=1, size=P.shape[1])

    y = np.zeros_like(x)
    rng = np.random.default_rng(seed)
    for i, Xi in enumerate(X):
        idxs = np.where(x == Xi)[0]
        y[idxs] = rng.choice(Y, size=idxs.shape, p=P[i])

    return convert_output(y)


@export
class Channel:
    """
    A base class for wireless channels.

    Group:
        simulation-channel-models
    """

    def __init__(self, seed: int | None = None):
        """
        Creates a new channel.

        Arguments:
            seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.
        """
        self._rng: np.random.Generator  # Will be set in self.reset()

        self.reset(seed)

    def reset(self, seed: int | None = None):
        """
        Resets the channel with a new seed.

        Arguments:
            seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.
        """
        self._rng = np.random.default_rng(seed)

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray:
        """
        Passes the input sequence $x$ through the channel.

        Arguments:
            x: The input sequence $x$.

        Returns:
            The output sequence $y$.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        return string

    @staticmethod
    def capacities() -> npt.NDArray:
        """
        Computes the channel capacity given the channel configuration.

        Returns:
            The channel capacities $C$ in bits/2D.
        """
        raise NotImplementedError

    @property
    def capacity(self) -> float:
        """
        The channel capacity $C$ in bits/2D of the instantiated channel.
        """
        raise NotImplementedError


@export
class BinarySymmetricChannel(Channel):
    r"""
    Implements a binary symmetric channel (BSC).

    Notes:
        The inputs to the BSC are $x_i \in \mathcal{X} = \{0, 1\}$ and the outputs are $y_i \in \mathcal{Y} = \{0, 1\}$.
        The capacity of the BSC is

        $$C = 1 - H_b(p) \ \ \text{bits/channel use} .$$

        This is an appropriate channel model for binary modulation with hard decisions at the detector.

    References:
        - John Proakis, *Digital Communications*, Chapter 6.5-1: Channel Models.

    Examples:
        When 20 bits are passed through a BSC with transition probability $p=0.25$, roughly 5 bits are flipped
        at the output.

        .. ipython:: python

            bsc = sdr.BinarySymmetricChannel(0.25, seed=1)
            x = np.random.randint(0, 2, 20); x
            y = bsc(x); y
            np.count_nonzero(x != y)

        The capacity of this BSC is 0.189 bits/channel use.

        .. ipython:: python

            bsc.capacity

        When the probability $p$ of bit error is 0, the capacity of the channel is 1 bit/channel use.
        However, as the probability of bit error approaches 0.5, the capacity of the channel approaches
        0.

        .. ipython:: python

            p = np.linspace(0, 1, 101); \
            C = sdr.BinarySymmetricChannel.capacities(p)

            @savefig sdr_BinarySymmetricChannel_1.png
            plt.figure(); \
            plt.plot(p, C); \
            plt.xlabel("Transition probability, $p$"); \
            plt.ylabel("Capacity (bits/channel use), $C$"); \
            plt.title("Capacity of the Binary Symmetric Channel");

    Group:
        simulation-channel-models
    """

    def __init__(self, p: float, seed: int | None = None):
        """
        Creates a new binary symmetric channel (BSC).

        Arguments:
            p: The transition probability $p$ of the BSC channel.
            seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.
        """
        self._p = verify_scalar(p, float=True, inclusive_min=0, inclusive_max=1)

        super().__init__(seed=seed)

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Passes the binary input sequence $x$ through the channel.

        Arguments:
            x: The input sequence $x$ with $x_i \in \mathcal{X} = \{0, 1\}$.

        Returns:
            The output sequence $y$ with $y_i \in \mathcal{Y} = \{0, 1\}$.
        """
        x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=2)

        flip = self._rng.choice([0, 1], size=x.shape, p=[1 - self.p, self.p])
        y = x ^ flip

        return convert_output(y)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(p={self.p})"

    def __str__(self) -> str:
        string = super().__str__()
        string += f" p={self.p}"
        return string

    @staticmethod
    def capacities(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """
        Calculates the capacity of BSC channels.

        Returns:
            The capacity $C$ of the channel in bits/channel use.
        """
        p = verify_arraylike(p, float=True, inclusive_min=0, inclusive_max=1)

        C = 1 - Hb(p)

        return convert_output(C)

    @property
    def X(self) -> npt.NDArray[np.int_]:
        r"""
        The input alphabet $\mathcal{X} = \{0, 1\}$ of the BSC channel.
        """
        return np.array([0, 1])

    @property
    def Y(self) -> npt.NDArray[np.int_]:
        r"""
        The output alphabet $\mathcal{Y} = \{0, 1\}$ of the BSC channel.
        """
        return np.array([0, 1])

    @property
    def p(self) -> float:
        """
        The transition probability $p$ of the BSC channel.
        """
        return self._p

    @property
    def capacity(self) -> float:
        """
        The capacity $C$ of the instantiated channel in bits/channel use.
        """
        return BinarySymmetricChannel.capacities(self.p)


@export
class BinaryErasureChannel(Channel):
    r"""
    Implements a binary erasure channel (BEC).

    Notes:
        The inputs to the BEC are $x_i \in \mathcal{X} = \{0, 1\}$ and the outputs are
        $y_i \in \mathcal{Y} = \{0, 1, e\}$. The capacity of the BEC is

        $$C = 1 - p \ \ \text{bits/channel use} .$$

    References:
        - John Proakis, *Digital Communications*, Chapter 6.5-1: Channel Models.

    Examples:
        When 20 bits are passed through a BEC with erasure probability $p=0.25$, roughly 5 bits are erased
        at the output.

        .. ipython:: python

            bec = sdr.BinaryErasureChannel(0.25, seed=1)
            x = np.random.randint(0, 2, 20); x
            y = bec(x); y
            np.count_nonzero(x != y)

        The capacity of this BEC is 0.75 bits/channel use.

        .. ipython:: python

            bec.capacity

        When the probability $p$ of bit erasure is 0, the capacity of the channel is 1 bit/channel use.
        However, as the probability of erasure approaches 1, the capacity of the channel linearly approaches
        0.

        .. ipython:: python

            p = np.linspace(0, 1, 101); \
            C = sdr.BinaryErasureChannel.capacities(p)

            @savefig sdr_BinaryErasureChannel_1.png
            plt.figure(); \
            plt.plot(p, C); \
            plt.xlabel("Erasure probability, $p$"); \
            plt.ylabel("Capacity (bits/channel use), $C$"); \
            plt.title("Capacity of the Binary Erasure Channel");

    Group:
        simulation-channel-models
    """

    def __init__(self, p: float, seed: int | None = None):
        """
        Creates a new binary erasure channel (BEC).

        Arguments:
            p: The erasure probability $p$ of the BEC channel.
            seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.
        """
        self._p = verify_scalar(p, float=True, inclusive_min=0, inclusive_max=1)

        super().__init__(seed=seed)

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Passes the binary input sequence $x$ through the channel.

        Arguments:
            x: The input sequence $x$ with $x_i \in \mathcal{X} = \{0, 1\}$.

        Returns:
            The output sequence $y$ with $y_i \in \mathcal{Y} = \{0, 1, e\}$. Erasures $e$ are represented by -1.
        """
        x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=2)

        random_p = self._rng.random(x.shape)
        y = np.where(random_p < self.p, -1, x)

        return convert_output(y)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(p={self.p})"

    def __str__(self) -> str:
        string = super().__str__()
        string += f" p={self.p}"
        return string

    @staticmethod
    def capacities(p: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """
        Calculates the capacity of BEC channels.

        Returns:
            The capacity $C$ of the channel in bits/channel use.
        """
        p = verify_arraylike(p, float=True, inclusive_min=0, inclusive_max=1)

        C = 1 - p

        return convert_output(C)

    @property
    def X(self) -> npt.NDArray[np.int_]:
        r"""
        The input alphabet $\mathcal{X} = \{0, 1\}$ of the BEC channel.
        """
        return np.array([0, 1])

    @property
    def Y(self) -> npt.NDArray[np.int_]:
        r"""
        The output alphabet $\mathcal{Y} = \{0, 1, e\}$ of the BEC channel. Erasures $e$ are represented by -1.
        """
        return np.array([0, 1, -1])

    @property
    def p(self) -> float:
        """
        The erasure probability $p$ of the BEC channel.
        """
        return self._p

    @property
    def capacity(self) -> float:
        """
        The capacity $C$ of the instantiated channel in bits/channel use.
        """
        return BinaryErasureChannel.capacities(self.p)


@export
class DiscreteMemorylessChannel(Channel):
    r"""
    Implements a discrete memoryless channel (DMC).

    Notes:
        The inputs to the DMC are $x_i \in \mathcal{X}$ and the outputs are $y_i \in \mathcal{Y}$.
        The capacity of the DMC is

        $$C = \max_{p(x)} I(X; Y) \ \ \text{bits/channel use} ,$$

        where $I(X; Y)$ is the mutual information between the input $X$ and output $Y$ of the channel.

    References:
        - John Proakis, *Digital Communications*, Chapter 6.5-1: Channel Models.

    Examples:
        Define the binary symmetric channel (BSC) with transition probability $p=0.25$.

        .. ipython:: python

            p = 0.25
            dmc = sdr.DiscreteMemorylessChannel([[1 - p, p], [p, 1 - p]], seed=1)
            dmc.P
            x = np.random.randint(0, 2, 20); x
            y = dmc(x); y
            np.count_nonzero(x != y)

        Define the binary erasure channel (BEC) with erasure probability $p=0.25$.

        .. ipython:: python

            p = 0.25
            dmc = sdr.DiscreteMemorylessChannel([[1 - p, 0, p], [0, 1 - p, p]], Y=[0, 1, -1], seed=1)
            x = np.random.randint(0, 2, 20); x
            y = dmc(x); y
            np.count_nonzero(x != y)

    Group:
        simulation-channel-models
    """

    def __init__(
        self,
        P: npt.ArrayLike,
        X: npt.ArrayLike | None = None,
        Y: npt.ArrayLike | None = None,
        seed: int | None = None,
    ):
        r"""
        Creates a new discrete memoryless channel (DMC).

        Arguments:
            P: The $m \times n$ transition probability matrix $P$, where $P = \Pr(Y = y_j | X = x_i)$.
            X: The input alphabet $\mathcal{X}$ of size $m$. If `None`, it is assumed that $\mathcal{X} = \{0, 1, \ldots, m-1\}$.
            Y: The output alphabet $\mathcal{Y}$ of size $n$. If `None`, it is assumed that $\mathcal{Y} = \{0, 1, \ldots, n-1\}$.
            seed: The seed for the random number generator. This is passed to :func:`numpy.random.default_rng`.
        """
        P = verify_arraylike(P, float=True, ndim=2, inclusive_min=0, inclusive_max=1)
        verify_condition(np.allclose(P.sum(axis=1), 1))
        self._P = P

        if X is None:
            X = np.arange(P.shape[0])
        else:
            X = verify_arraylike(X, int=True, ndim=1, size=P.shape[0])
        self._X = X

        if Y is None:
            Y = np.arange(P.shape[1])
        else:
            Y = verify_arraylike(Y, int=True, ndim=1, size=P.shape[1])
        self._Y = Y

        super().__init__(seed=seed)

    def __call__(self, x: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Passes the input sequence $x$ through the channel.

        Arguments:
            x: The input sequence $x$ with $x_i \in \mathcal{X}$.

        Returns:
            The output sequence $y$ with $y_i \in \mathcal{Y}$.
        """
        x = verify_arraylike(x, int=True, inclusive_min=0, exclusive_max=len(self.X))

        y = np.zeros_like(x)
        for i, Xi in enumerate(self.X):
            idxs = np.where(x == Xi)[0]
            y[idxs] = self._rng.choice(self.Y, size=idxs.shape, p=self.P[i])

        return convert_output(y)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(P={self.P}, X={self.X}, Y={self.Y})"

    def __str__(self) -> str:
        string = super().__str__()
        string += f" P={self.P}, X={self.X}, Y={self.Y}"
        return string

    @property
    def X(self) -> npt.NDArray[np.int_]:
        r"""
        The input alphabet $\mathcal{X}$ of the DMC channel.
        """
        return self._X

    @property
    def Y(self) -> npt.NDArray[np.int_]:
        r"""
        The output alphabet $\mathcal{Y}$ of the DMC channel.
        """
        return self._Y

    @property
    def P(self) -> npt.NDArray[np.float64]:
        """
        The transition probability matrix $P$ of the DMC channel.
        """
        return self._P



"""
A module containing functions for data manipulation.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from ._helper import convert_output, export, verify_arraylike, verify_scalar


@export
def pack(
    x: npt.ArrayLike,
    bpe: int,
    dtype: npt.DTypeLike | None = None,
) -> npt.NDArray[np.int_]:
    """
    Packs a binary array into an array with multiple bits per element.

    The data is assumed to have the most significant bit first. If there are not enough bits in the
    input array to fill the last element of the output array, the remaining bits are filled with
    zeros.

    Arguments:
        x: The input binary array with 1 bit per element.
        bpe: The number of bits per element in the output array.
        dtype: The data type of the output array. If `None`, the smallest unsigned integer dtype
            that can hold `bpe` bits is used.

    Returns:
        The packed data with `bpe` bits per element.

    Examples:
        .. ipython:: python

            sdr.pack([1, 0, 0, 0, 1, 1, 0, 1], 2)
            sdr.pack([1, 0, 0, 0, 1, 1, 0, 1], 3)

    Group:
        data
    """
    x = verify_arraylike(x, int=True, atleast_1d=True)
    verify_scalar(bpe, int=True, positive=True)

    if dtype is None:
        dtype = _unsigned_dtype(bpe)

    if bpe == 1:
        y = x
    elif bpe == 8:
        y = np.packbits(x, axis=-1)
    else:
        if x.shape[-1] % bpe != 0:
            # Pad the last axis
            pad = bpe - (x.shape[-1] % bpe)
            pad_width = ((0, 0),) * (x.ndim - 1) + ((0, pad),)
            x = np.pad(x, pad_width, mode="constant", constant_values=0)

        shape = x.shape[:-1] + (x.shape[-1] // bpe, bpe)
        X = x.reshape(shape)
        single_bits = 2 ** np.arange(bpe - 1, -1, -1, dtype=dtype)
        y = np.sum(X * single_bits, axis=-1)

    y = y.astype(dtype)

    return convert_output(y)


@export
def unpack(
    x: npt.ArrayLike,
    bpe: int,
    dtype: npt.DTypeLike | None = None,
) -> npt.NDArray[np.int_]:
    """
    Unpacks an array with multiple bits per element into a binary array.

    The data is assumed to have the most significant bit first.

    Arguments:
        x: The input array with `bpe` bits per element.
        bpe: The number of bits per element in the input array.
        dtype: The data type of the output array. If `None`, :obj:`numpy.uint8` is used.

    Returns:
        The unpacked binary data with 1 bit per element.

    Examples:
        .. ipython:: python

            sdr.unpack([2, 0, 3, 1], 2)
            sdr.unpack([4, 3, 2], 3)

    Group:
        data
    """
    x = verify_arraylike(x, int=True, atleast_1d=True)
    verify_scalar(bpe, int=True, positive=True)

    if dtype is None:
        dtype = np.uint8

    if bpe == 1:
        y = x
    elif bpe == 8:
        # The unpackbits() function requires unsigned dtypes.
        dtype = _unsigned_dtype(bpe)
        y = np.unpackbits(x.astype(dtype), axis=-1)
    else:
        single_bits = 2 ** np.arange(bpe - 1, -1, -1, dtype=np.int64)
        X = (x[..., np.newaxis] & single_bits) > 0
        shape = x.shape[:-1] + (x.shape[-1] * bpe,)
        y = X.reshape(shape)

    y = y.astype(dtype)

    return convert_output(y)


@export
def hexdump(
    data: npt.ArrayLike | bytes,
    width: int = 16,
) -> str:
    """
    Returns a hexdump of the specified data.

    Arguments:
        data: The data to display. Each element is considered one byte. Use :func:`sdr.pack()`
            or :func:`sdr.unpack` to convert data with variable bits per element.
        width: The number of bytes per line.

    Returns:
        A string containing the hexdump of the data.

    Examples:
        .. ipython:: python

            print(sdr.hexdump(b"The quick brown fox jumps over the lazy dog"))

        .. ipython:: python

            print(sdr.hexdump([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], width=4))

    Group:
        data
    """
    if isinstance(data, bytes):
        data = np.frombuffer(data, dtype=np.uint8)
    else:
        data = verify_arraylike(data, int=True, ndim=1, inclusive_min=0, exclusive_max=256)
    verify_scalar(width, int=True, inclusive_min=1, inclusive_max=16)

    if width > data.size:
        width = data.size

    string = ""
    for i in range(0, data.size, width):
        line = data[i : i + width]
        string += f"{i:08x}  "
        string += " ".join(f"{x:02x}" for x in line)
        string += "   " * (width - line.size)
        string += "  "
        string += "".join(chr(x) if 32 <= x <= 126 else "." for x in line)
        string += "\n"

    return string


def _unsigned_dtype(bits: int) -> np.integer:
    """
    Returns the smallest unsigned integer dtype that can hold the specified number of bits.
    """
    if bits <= 8:
        return np.uint8
    if bits <= 16:
        return np.uint16
    if bits <= 32:
        return np.uint32
    if bits <= 64:
        return np.uint64
    if bits <= 128:
        return np.uint128
    if bits <= 256:
        return np.uint256

    raise ValueError(f"Cannot create unsigned integer dtype for {bits} bits.")



"""
A module containing functions for signal manipulation.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.signal

from ._helper import convert_output, export, verify_arraylike, verify_bool, verify_scalar


@export
def mix(
    x: npt.ArrayLike,
    freq: float = 0.0,
    phase: float = 0.0,
    sample_rate: float = 1.0,
    complex: bool = True,
) -> npt.NDArray:
    r"""
    Mixes a time-domain signal with a complex exponential or real sinusoid.

    Arguments:
        x: The time-domain signal $x[n]$.
        freq: The frequency $f$ of the sinusoid in Hz (or 1/samples if `sample_rate=1`).
            The frequency must satisfy $-f_s/2 \le f \le f_s/2$.
        phase: The phase $\phi$ of the sinusoid in degrees.
        sample_rate: The sample rate $f_s$ of the signal.
        complex: Indicates whether to mix by a complex exponential or real sinusoid.

            - `True`: $y[n] = x[n] \cdot \exp \left[ j \left( \frac{2 \pi f}{f_s} n + \phi \right) \right]$
            - `False`: $y[n] = x[n] \cdot \cos \left( \frac{2 \pi f}{f_s} n + \phi \right)$

    Returns:
        The mixed signal $y[n]$.

    Examples:
        Create a complex exponential with a frequency of 10 Hz and phase of 45 degrees.

        .. ipython:: python

            sample_rate = 1e3; \
            N = 100; \
            x = np.exp(1j * (2 * np.pi * 10 * np.arange(N) / sample_rate + np.pi/4))

            @savefig sdr_mix_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=sample_rate); \
            plt.title(r"Complex exponential with $f=10$ Hz and $\phi=45$ degrees");

        Mix the signal to baseband by removing the frequency rotation and the phase offset.

        .. ipython:: python

            y = sdr.mix(x, freq=-10, phase=-45, sample_rate=sample_rate)

            @savefig sdr_mix_2.png
            plt.figure(); \
            sdr.plot.time_domain(y, sample_rate=sample_rate); \
            plt.title(r"Baseband signal with $f=0$ Hz and $\phi=0$ degrees");

    Group:
        dsp-signal-manipulation
    """
    x = verify_arraylike(x, ndim=1)
    verify_scalar(freq, float=True, inclusive_min=-sample_rate / 2, inclusive_max=sample_rate / 2)
    verify_scalar(phase, float=True)
    verify_scalar(sample_rate, float=True, positive=True)
    verify_bool(complex)

    t = np.arange(x.size) / sample_rate  # Time vector in seconds
    if complex:
        lo = np.exp(1j * (2 * np.pi * freq * t + np.deg2rad(phase)))
    else:
        lo = np.cos(2 * np.pi * freq * t + np.deg2rad(phase))

    y = x * lo

    return convert_output(y)


@export
def to_complex_baseband(x_r: npt.ArrayLike) -> npt.NDArray[np.complex128]:
    r"""
    Converts a real passband signal to a complex baseband signal.

    The real passband signal $x_r[n]$ is centered at $f_{s,r}/4$ with sample rate $f_{s,r}$.
    The complex baseband signal $x_c[n]$ is centered at $0$ with sample rate $f_{s,c} = f_{s,r}/2$.

    Arguments:
        x_r: The real passband signal $x_r[n]$ centered at $f_{s,r}/4$ with sample rate $f_{s,r}$.
            If the length is odd, one zero is appended to the end.

    Returns:
        The complex baseband signal $x_c[n]$ centered at $0$ with sample rate $f_{s,c} = f_{s,r}/2$.

    Examples:
        Create a real passband signal with frequency components at 100, 250, and 300 Hz, at a sample rate of 1 ksps.
        Notice the spectrum is complex-conjugate symmetric.

        .. ipython:: python

            sample_rate = 1e3; \
            x_r = ( \
                0.1 * np.sin(2 * np.pi * 100 / sample_rate * np.arange(1000)) \
                + 1.0 * np.sin(2 * np.pi * 250 / sample_rate * np.arange(1000)) \
                + 0.5 * np.sin(2 * np.pi * 300 / sample_rate * np.arange(1000)) \
            ); \
            x_r = sdr.awgn(x_r, snr=30)

            @savefig sdr_to_complex_baseband_1.png
            plt.figure(); \
            sdr.plot.time_domain(x_r[0:100], sample_rate=sample_rate); \
            plt.title("Time-domain signal $x_r[n]$");

            @savefig sdr_to_complex_baseband_2.png
            plt.figure(); \
            sdr.plot.periodogram(x_r, fft=2048, sample_rate=sample_rate); \
            plt.title("Periodogram of $x_r[n]$");

        Convert the real passband signal to a complex baseband signal with sample rate 500 sps and center of 0 Hz.
        Notice the spectrum is no longer complex-conjugate symmetric.
        The real sinusoids are now complex exponentials at -150, 0, and 50 Hz.

        .. ipython:: python

            x_c = sdr.to_complex_baseband(x_r); \
            sample_rate /= 2

            @savefig sdr_to_complex_baseband_3.png
            plt.figure(); \
            sdr.plot.time_domain(x_c[0:50], sample_rate=sample_rate); \
            plt.title("Time-domain signal $x_c[n]$");

            @savefig sdr_to_complex_baseband_4.png
            plt.figure(); \
            sdr.plot.periodogram(x_c, fft=2048, sample_rate=sample_rate); \
            plt.title("Periodogram of $x_c[n]$");

    Group:
        dsp-signal-manipulation
    """
    x_r = verify_arraylike(x_r, float=True, real=True, ndim=1)

    if x_r.size % 2 == 1:
        # Append one zero for the decimate by 2
        x_r = np.append(x_r, 0)

    # Compute the analytic signal. The analytic signal has the negative frequencies zeroed out.
    x_a = scipy.signal.hilbert(x_r)

    # Mix the complex signal to baseband. What was fs/4 is now 0.
    x_c = mix(x_a, freq=-0.25)

    # Decimate by 2 to achieve fs/2 sample rate. This can be done without anti-alias filtering since the
    # spectrum outside +/- fs/4 is zero.
    x_c = x_c[::2]

    return convert_output(x_c)


@export
def to_real_passband(x_c: npt.ArrayLike) -> npt.NDArray[np.float64]:
    r"""
    Converts a complex baseband signal to a real passband signal.

    The complex baseband signal $x_c[n]$ is centered at $0$ with sample rate $f_{s,c}$.
    The real passband signal $x_r[n]$ is centered at $f_{s,r}/4$ with sample rate $f_{s,r} = 2f_{s,c}$.

    Arguments:
        x_c: The complex baseband signal $x_c[n]$ centered at $0$ with sample rate $f_{s,c}$.

    Returns:
        The real passband signal $x_r[n]$ centered at $f_{s,r}/4$ with sample rate $f_{s,r} = 2f_{s,c}$.

    Examples:
        Create a complex baseband signal with frequency components at -150, 0, and 50 Hz, at a sample rate of 500 sps.
        Notice the spectrum is asymmetric.

        .. ipython:: python

            sample_rate = 500; \
            x_c = ( \
                0.1 * np.exp(1j * 2 * np.pi * -150 / sample_rate * np.arange(1000)) \
                + 1.0 * np.exp(1j * 2 * np.pi * 0 / sample_rate * np.arange(1000)) \
                + 0.5 * np.exp(1j * 2 * np.pi * 50 / sample_rate * np.arange(1000)) \
            ); \
            x_c = sdr.awgn(x_c, snr=30)

            @savefig sdr_to_real_passband_1.png
            plt.figure(); \
            sdr.plot.time_domain(x_c[0:50], sample_rate=sample_rate); \
            plt.title("Time-domain signal $x_c[n]$");

            @savefig sdr_to_real_passband_2.png
            plt.figure(); \
            sdr.plot.periodogram(x_c, fft=2048, sample_rate=sample_rate); \
            plt.title("Periodogram of $x_c[n]$");

        Convert the complex baseband signal to a real passband signal with sample rate 1 ksps and center of 250 Hz.
        Notice the spectrum is now complex-conjugate symmetric.
        The complex exponentials are now real sinusoids at 100, 250, and 300 Hz.

        .. ipython:: python

            x_r = sdr.to_real_passband(x_c); \
            sample_rate *= 2

            @savefig sdr_to_real_passband_3.png
            plt.figure(); \
            sdr.plot.time_domain(x_r[0:100], sample_rate=sample_rate); \
            plt.title("Time-domain signal $x_r[n]$");

            @savefig sdr_to_real_passband_4.png
            plt.figure(); \
            sdr.plot.periodogram(x_r, fft=2048, sample_rate=sample_rate); \
            plt.title("Periodogram of $x_r[n]$");

    Group:
        dsp-signal-manipulation
    """
    x_c = verify_arraylike(x_c, complex=True, imaginary=True, ndim=1)

    # Upsample by 2 to achieve 2*fs sample rate
    x_c = scipy.signal.resample_poly(x_c, 2, 1)

    # Mix the complex baseband signal to passband. What was 0 is now fs/4.
    x_c = mix(x_c, freq=0.25)

    # Only preserve the real part, which is complex-conjugate symmetric about 0 Hz.
    x_r = x_c.real

    return convert_output(x_r)


@export
def upsample(x: npt.ArrayLike, rate: int) -> npt.NDArray:
    r"""
    Upsamples the time-domain signal $x[n]$ by the factor $r$, by inserting $r-1$ zeros between each sample.

    Warning:
        This function does not perform any anti-aliasing filtering. The upsampled signal $y[n]$ will have
        frequency components above the Nyquist frequency of the original signal $x[n]$. For efficient
        polyphase interpolation (with anti-aliasing filtering), see :class:`sdr.Interpolator`.

    Arguments:
        x: The time-domain signal $x[n]$ with sample rate $f_s$.
        rate: The upsampling factor $r$.

    Returns:
        The upsampled signal $y[n]$ with sample rate $f_s r$.

    See Also:
        sdr.Interpolator

    Examples:
        Upsample a complex exponential by a factor of 4.

        .. ipython:: python

            sample_rate = 100; \
            x = np.exp(1j * 2 * np.pi * 15 / sample_rate * np.arange(20))
            y = sdr.upsample(x, 4)

            @savefig sdr_upsample_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=sample_rate); \
            plt.title("Input signal $x[n]$");

            @savefig sdr_upsample_2.png
            plt.figure(); \
            sdr.plot.time_domain(y, sample_rate=sample_rate*4); \
            plt.title("Upsampled signal $y[n]$");

        The spectrum of $y[n]$ has 3 additional copies of the spectrum of $x[n]$.

        .. ipython:: python
            :okwarning:

            @savefig sdr_upsample_3.png
            plt.figure(); \
            sdr.plot.periodogram(x, fft=2048, sample_rate=sample_rate); \
            plt.xlim(-sample_rate*2, sample_rate*2); \
            plt.ylim(-100, 0); \
            plt.title("Input signal $x[n]$");

            @savefig sdr_upsample_4.png
            plt.figure(); \
            sdr.plot.periodogram(y, fft=2048, sample_rate=sample_rate*4); \
            plt.xlim(-sample_rate*2, sample_rate*2); \
            plt.ylim(-100, 0); \
            plt.title("Upsampled signal $y[n]$");

    Group:
        dsp-signal-manipulation
    """
    x = verify_arraylike(x, ndim=1)
    verify_scalar(rate, int=True, positive=True)

    y = np.zeros(x.size * rate, dtype=x.dtype)
    y[::rate] = x

    return convert_output(y)


@export
def downsample(x: npt.ArrayLike, rate: int) -> npt.NDArray:
    r"""
    Downsamples the time-domain signal $x[n]$ by the factor $r$, by discarding $r-1$ samples every $r$ samples.

    Warning:
        This function does not perform any anti-aliasing filtering. The downsampled signal $y[n]$ will have
        spectral aliasing. For efficient polyphase decimation (with anti-aliasing filtering), see
        :class:`sdr.Decimator`.

    Arguments:
        x: The time-domain signal $x[n]$ with sample rate $f_s$.
        rate: The downsampling factor $r$.

    Returns:
        The downsampled signal $y[n]$ with sample rate $f_s / r$.

    Examples:
        Downsample a complex exponential by a factor of 4.

        .. ipython:: python

            sample_rate = 400; \
            x1 = np.exp(1j * 2 * np.pi * 0 / sample_rate * np.arange(200)); \
            x2 = np.exp(1j * 2 * np.pi * 130 / sample_rate * np.arange(200)); \
            x3 = np.exp(1j * 2 * np.pi * -140 / sample_rate * np.arange(200)); \
            x = x1 + x2 + x3
            y = sdr.downsample(x, 4)

            @savefig sdr_downsample_1.png
            plt.figure(); \
            sdr.plot.time_domain(x, sample_rate=sample_rate); \
            plt.title("Input signal $x[n]$");

            @savefig sdr_downsample_2.png
            plt.figure(); \
            sdr.plot.time_domain(y, sample_rate=sample_rate/4); \
            plt.title("Downsampled signal $y[n]$");

        The spectrum of $x[n]$ has aliased. Any spectral content above the Nyquist frequency of $f_s / 2$
        will *fold* into the spectrum of $y[n]$. The CW at 0 Hz remains at 0 Hz (unaliased).
        The CW at 130 Hz folds into 30 Hz. The CW at -140 Hz folds into -40 Hz.

        .. ipython:: python
            :okwarning:

            @savefig sdr_downsample_3.png
            plt.figure(); \
            sdr.plot.periodogram(x, fft=2048, sample_rate=sample_rate); \
            plt.xlim(-sample_rate/2, sample_rate/2); \
            plt.ylim(-100, 0); \
            plt.title("Input signal $x[n]$");

            @savefig sdr_downsample_4.png
            plt.figure(); \
            sdr.plot.periodogram(y, fft=2048, sample_rate=sample_rate/4); \
            plt.xlim(-sample_rate/2, sample_rate/2); \
            plt.ylim(-100, 0); \
            plt.title("Downsampled signal $y[n]$");

    Group:
        dsp-signal-manipulation
    """
    x = verify_arraylike(x, ndim=1)
    verify_scalar(rate, int=True, positive=True)

    y = x[::rate]

    return convert_output(y)



import numpy as np
import pytest

import sdr


def test_constant():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.ones(N)
    assert sdr.energy(x) == pytest.approx(A**2 * N, rel=1e-2)


def test_sine():
    rng = np.random.default_rng()
    A = rng.uniform(2, 10)  # Signal amplitude
    N = 50
    x = A * np.sin(np.linspace(0, 2 * np.pi, N, endpoint=False))
    assert sdr.energy(x) == pytest.approx(A**2 / 2 * N, rel=1e-2)



import numpy as np

import sdr

X = np.array([[6, 5, 0, 0, 5], [0, 1, 5, 0, 3], [4, 3, 3, 0, 5]])
Y = np.array([[7, 5, 2, 4, 0], [0, 0, 4, 1, 5], [5, 0, 7, 5, 6]])


def test_axis_none():
    d = sdr.hamming(X, Y)
    assert d == 37


def test_axis_0():
    d = sdr.hamming(X, Y, axis=0)
    assert d.shape == (5,)
    assert np.allclose(d, [2, 4, 7, 10, 14])


def test_axis_1():
    d = sdr.hamming(X, Y, axis=1)
    assert d.shape == (3,)
    assert np.allclose(d, [12, 9, 16])



"""
A module containing functions for various symbol coding schemes.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from .._helper import convert_output, export, verify_arraylike, verify_scalar


@export
def diff_encode(
    x: npt.ArrayLike,
    y_prev: int = 0,
) -> npt.NDArray[np.int_]:
    r"""
    Differentially encodes the input data $x[k]$.

    $$y[k] = x[k] \oplus y[k-1]$$

    Args:
        x: The input uncoded data $x[k]$.
        y_prev: The previous value of the output encoded data $y[k-1]$.

    Returns:
        The differentially encoded data $y[k]$.

    Notes:
        .. code-block:: text
            :caption: Differential Encoder Block Diagram

            x[k] -->@--------------+--> y[k]
                    ^              |
                    |   +------+   |
            y[k-1]  +---| z^-1 |<--+
                        +------+

            x[k] = Input data
            y[k] = Encoded data
            z^-1 = Unit delay
            @ = Adder

    Examples:
        .. ipython:: python

            sdr.diff_encode([0, 1, 0, 0, 1, 1])
            sdr.diff_decode([0, 1, 1, 1, 0, 1])

    Group:
        modulation-symbol-encoding
    """
    x = verify_arraylike(x, int=True, atleast_1d=True, ndim=1, non_negative=True)
    verify_scalar(y_prev, int=True, non_negative=True)

    y = np.empty_like(x, dtype=int)
    y[0] = x[0] ^ y_prev
    for n in range(1, len(x)):
        y[n] = x[n] ^ y[n - 1]

    return convert_output(y)


@export
def diff_decode(
    y: npt.ArrayLike,
    y_prev: int = 0,
) -> npt.NDArray[np.int_]:
    r"""
    Differentially decodes the input data $y[k]$.

    $$x[k] = y[k] \oplus y[k-1]$$

    Arguments:
        y: The input encoded data $y[k]$.
        y_prev: The previous value of the encoded data $y[k-1]$.

    Returns:
        The differentially decoded data $x[k]$.

    Notes:
        .. code-block:: text
            :caption: Differential Decoder Block Diagram

                       +------+  y[k-1]
            y[k] --+-->| z^-1 |--------->@--> x[k]
                   |   +------+          ^
                   |                     |
                   +---------------------+

            y[k] = Encoded data
            x[k] = Decoded data
            z^-1 = Unit delay
            @ = Adder

    Examples:
        .. ipython:: python

            sdr.diff_decode([0, 1, 1, 1, 0, 1])
            sdr.diff_encode([0, 1, 0, 0, 1, 1])

    Group:
        modulation-symbol-encoding
    """
    y = verify_arraylike(y, int=True, atleast_1d=True, ndim=1, non_negative=True)
    verify_scalar(y_prev, int=True, non_negative=True)

    x = np.empty_like(y, dtype=int)
    x[0] = y[0] ^ y_prev
    for n in range(1, len(y)):
        x[n] = y[n] ^ y[n - 1]

    return convert_output(x)



"""
A module containing classes for phase-shift keying (PSK) modulations.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.integrate
import scipy.special
from typing_extensions import Literal

from .._conversion import ebn0_to_esn0, esn0_to_ebn0, linear
from .._data import unpack
from .._helper import (
    convert_output,
    export,
    extend_docstring,
    verify_arraylike,
    verify_bool,
    verify_condition,
    verify_literal,
    verify_scalar,
)
from .._probability import Q
from .._sequence import binary_code, gray_code
from ._linear import LinearModulation


@export
class PSK(LinearModulation):
    r"""
    Implements phase-shift keying (PSK) modulation and demodulation.

    Notes:
        Phase-shift keying (PSK) is a linear phase modulation scheme that encodes information by modulating
        the phase of a carrier sinusoid. The modulation order $M = 2^k$ is a power of 2 and indicates the number of
        phases used. The input bit stream is taken $k$ bits at a time to create decimal symbols
        $s[k] \in \{0, \dots, M-1\}$. These decimal symbols $s[k]$ are then mapped to complex symbols
        $a[k] \in \mathbb{C}$ by the equation

        $$a[k] = \exp \left[ j\left(\frac{2\pi}{M}s[k] + \phi\right) \right] .$$

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Examples:
        Create a QPSK modem whose constellation has a 45° phase offset.

        .. ipython:: python

            qpsk = sdr.PSK(4, phase_offset=45, pulse_shape="srrc"); qpsk

            @savefig sdr_PSK_1.png
            plt.figure(); \
            sdr.plot.symbol_map(qpsk);

        Generate a random bit stream, convert to 2-bit symbols, and map to complex symbols.

        .. ipython:: python

            bits = np.random.randint(0, 2, 1000); bits[0:8]
            symbols = sdr.pack(bits, qpsk.bps); symbols[0:4]
            complex_symbols = qpsk.map_symbols(symbols); complex_symbols[0:4]

            @savefig sdr_PSK_2.png
            plt.figure(); \
            sdr.plot.constellation(complex_symbols, linestyle="-");

        Modulate and pulse shape the symbols to a complex baseband signal.

        .. ipython:: python

            tx_samples = qpsk.modulate(symbols)

            @savefig sdr_PSK_3.png
            plt.figure(); \
            sdr.plot.time_domain(tx_samples[0:50*qpsk.sps]);

        Examine the eye diagram of the pulse-shaped transmitted signal. The SRRC pulse shape is not a Nyquist filter,
        so ISI is present.

        .. ipython:: python

            @savefig sdr_PSK_4.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[5*qpsk.sps : -5*qpsk.sps], qpsk.sps, persistence=True); \
            plt.suptitle("Noiseless transmitted signal with ISI");

        Add AWGN noise such that $E_b/N_0 = 30$ dB.

        .. ipython:: python

            ebn0 = 30; \
            snr = sdr.ebn0_to_snr(ebn0, bps=qpsk.bps, sps=qpsk.sps); \
            rx_samples = sdr.awgn(tx_samples, snr=snr)

            @savefig sdr_PSK_5.png
            plt.figure(); \
            sdr.plot.time_domain(rx_samples[0:50*qpsk.sps]);

        Manually apply a matched filter. Examine the eye diagram of the matched filtered received signal. The
        two cascaded SRRC filters create a Nyquist RC filter. Therefore, the ISI is removed.

        .. ipython:: python

            mf = sdr.FIR(qpsk.pulse_shape); \
            mf_samples = mf(rx_samples)

            @savefig sdr_PSK_6.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(mf_samples[10*qpsk.sps : -10*qpsk.sps], qpsk.sps, persistence=True); \
            plt.suptitle("Noisy received and matched filtered signal without ISI");

        Matched filter and demodulate.

        .. ipython:: python

            rx_symbols, rx_complex_symbols, _ = qpsk.demodulate(rx_samples)

            # The symbol decisions are error-free
            np.array_equal(symbols, rx_symbols)

            @savefig sdr_PSK_7.png
            plt.figure(); \
            sdr.plot.constellation(rx_complex_symbols);

        See the :ref:`psk` example.

    Group:
        modulation-linear
    """

    def __init__(
        self,
        order: int,
        phase_offset: float = 0.0,
        symbol_labels: Literal["bin", "gray"] | npt.ArrayLike = "gray",
        sps: int = 8,
        pulse_shape: npt.ArrayLike | Literal["rect", "rc", "srrc"] = "rect",
        span: int | None = None,
        alpha: float | None = None,
    ):
        r"""
        Creates a new PSK object.

        Arguments:
            order: The modulation order $M = 2^k$, where $k \ge 1$ is the coded bits per symbol.
            phase_offset: The phase offset $\phi$ in degrees.
            symbol_labels: The decimal symbol labels of consecutive complex symbols.

                - `"bin"`: The symbols are binary-coded. Adjacent symbols may differ by more than one bit.
                - `"gray":` The symbols are Gray-coded. Adjacent symbols only differ by one bit.
                - `npt.ArrayLike`: An $M$-length array whose indices are the default symbol labels and whose values are
                  the new symbol labels. The default symbol labels are $0$ to $M-1$ for phases starting at $1 + 0j$
                  and going counter-clockwise around the unit circle.

            sps: The number of samples per symbol $f_s / f_{sym}$.
            pulse_shape: The pulse shape $h[n]$ of the modulated signal.

                - `npt.ArrayLike`: A custom pulse shape. It is important that `sps` matches the design
                  of the pulse shape. See :ref:`pulse-shaping-functions`.
                - `"rect"`: Rectangular pulse shape.
                - `"rc"`: Raised cosine pulse shape.
                - `"srrc"`: Square-root raised cosine pulse shape.

            span: The span of the pulse shape in symbols. This is only used if `pulse_shape` is a string.
                If `None`, 1 is used for `"rect"` and 10 is used for `"rc"` and `"srrc"`.
            alpha: The roll-off factor of the pulse shape. If `None`, 0.2 is used for `"rc"` and `"srrc"`.

        See Also:
            sdr.rectangular, sdr.raised_cosine, sdr.root_raised_cosine
        """
        verify_scalar(order, int=True, positive=True, power_of_two=True)
        verify_scalar(phase_offset, float=True)

        # Define the base PSK symbol map
        base_symbol_map = np.exp(1j * (2 * np.pi * np.arange(order) / order + np.deg2rad(phase_offset)))

        super().__init__(
            base_symbol_map,
            phase_offset=phase_offset,
            sps=sps,
            pulse_shape=pulse_shape,
            span=span,
            alpha=alpha,
        )

        if isinstance(symbol_labels, str):
            verify_literal(symbol_labels, ["bin", "gray"])
            if symbol_labels == "bin":
                self._symbol_labels = binary_code(self.order)
                self._symbol_labels_str = "bin"
            elif symbol_labels == "gray":
                self._symbol_labels = gray_code(self.order)
                self._symbol_labels_str = "gray"
        else:
            symbol_labels = verify_arraylike(symbol_labels, int=True, ndim=1, size=self.order)
            verify_condition(np.unique(symbol_labels).size == self.order)
            self._symbol_labels = symbol_labels
            self._symbol_labels_str = self._symbol_labels

        # Relabel the symbols
        self._symbol_map[self._symbol_labels] = self._symbol_map.copy()

    def __repr__(self) -> str:
        return (
            f"sdr.{type(self).__name__}({self.order}, phase_offset={self.phase_offset}, "
            + f"symbol_labels={self._symbol_labels_str!r})"
        )

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  order: {self.order}"
        string += f"\n  symbol_map: {self.symbol_map.shape} shape"
        string += f"\n    {self.symbol_map.tolist()}"
        string += f"\n  symbol_labels: {self._symbol_labels_str!r}"
        string += f"\n  phase_offset: {self.phase_offset}"
        return string

    def ber(
        self,
        ebn0: npt.ArrayLike,
        diff_encoded: bool = False,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the bit error rate (BER) at the provided $E_b/N_0$ values.

        Arguments:
            ebn0: Bit energy $E_b$ to noise PSD $N_0$ ratio in dB.
            diff_encoded: Indicates whether the input symbols were differentially encoded.

        Returns:
            The bit error rate $P_{be}$.

        See Also:
            sdr.esn0_to_ebn0, sdr.snr_to_ebn0

        References:
            - Simon and Alouini, *Digital Communications over Fading Channels*,
              Chapter 8: Performance of Single-Channel Receivers.
            - John Proakis, *Digital Communications*, Chapter 4: Optimum Receivers for AWGN Channels.

        Examples:
            See the :ref:`psk` example.

            Plot theoretical BER curves for BPSK, QPSK, 8-PSK, and 16-PSK in an AWGN channel.

            .. ipython:: python

                bpsk = sdr.PSK(2); \
                qpsk = sdr.PSK(4); \
                psk8 = sdr.PSK(8); \
                psk16 = sdr.PSK(16); \
                ebn0 = np.linspace(-2, 10, 100)

                @savefig sdr_psk_ber_1.png
                plt.figure(); \
                sdr.plot.ber(ebn0, bpsk.ber(ebn0), label="BPSK"); \
                sdr.plot.ber(ebn0, qpsk.ber(ebn0), label="QPSK"); \
                sdr.plot.ber(ebn0, psk8.ber(ebn0), label="8-PSK"); \
                sdr.plot.ber(ebn0, psk16.ber(ebn0), label="16-PSK"); \
                plt.title("BER curves for PSK modulation in an AWGN channel");

            Compare the bit error rate of QPSK and DE-QPSK in an AWGN channel.

            .. ipython:: python

                @savefig sdr_psk_ber_2.png
                plt.figure(); \
                sdr.plot.ber(ebn0, qpsk.ber(ebn0), label="QPSK"); \
                sdr.plot.ber(ebn0, qpsk.ber(ebn0, diff_encoded=True), label="DE-QPSK"); \
                plt.title("BER curves for PSK and DE-PSK modulation in an AWGN channel");
        """
        ebn0 = verify_arraylike(ebn0, float=True)
        verify_bool(diff_encoded)

        M = self.order
        k = self.bps
        ebn0_linear = linear(ebn0)
        esn0 = ebn0_to_esn0(ebn0, k)
        esn0_linear = linear(esn0)

        if not diff_encoded:
            if M == 2:
                # Equation 4.3-13 from Proakis
                Pbe = Q(np.sqrt(2 * ebn0_linear))
            elif M == 4 and np.array_equal(self._symbol_labels, gray_code(M)):
                # Equation 4.3-13 from Proakis
                Pbe = Q(np.sqrt(2 * ebn0_linear))
            else:
                # Equation 8.29 from Simon and Alouini
                Pbe = np.zeros_like(esn0_linear)
                for i in range(esn0_linear.size):
                    for j in range(1, M):
                        Pj = Pk(M, esn0_linear[i], j)
                        # The number of bits that differ between symbol j and symbol 0
                        N_bits = unpack(self._symbol_labels[j] ^ self._symbol_labels[0], k).sum()
                        Pbe[i] += Pj * N_bits
                    # Equation 8.31 from Simon and Alouini
                    Pbe[i] /= k
        else:
            if M == 2:
                # Equation 8.37 from Simon and Alouini
                Pbe = 2 * Q(np.sqrt(2 * ebn0_linear)) - 2 * Q(np.sqrt(2 * ebn0_linear)) ** 2
            elif M == 4 and np.array_equal(self._symbol_labels, gray_code(M)):
                # Equation 8.37 from Simon and Alouini
                Pbe = 2 * Q(np.sqrt(2 * ebn0_linear)) - 2 * Q(np.sqrt(2 * ebn0_linear)) ** 2
            else:
                raise ValueError("Differential encoding is not supported for M-PSK with M > 4.")

        return convert_output(Pbe)

    def ser(
        self,
        esn0: npt.ArrayLike,
        diff_encoded: bool = False,
    ) -> npt.NDArray[np.float64]:
        r"""
        Computes the symbol error rate (SER) at the provided $E_s/N_0$ values.

        Arguments:
            esn0: Symbol energy $E_s$ to noise PSD $N_0$ ratio in dB.
            diff_encoded: Indicates whether the input symbols were differentially encoded.

        Returns:
            The symbol error rate $P_{se}$.

        See Also:
            sdr.ebn0_to_esn0, sdr.snr_to_esn0

        References:
            - Simon and Alouini, *Digital Communications over Fading Channels*,
              Chapter 8: Performance of Single-Channel Receivers.
            - John Proakis, *Digital Communications*, Chapter 4: Optimum Receivers for AWGN Channels.

        Examples:
            See the :ref:`psk` example.

            Plot theoretical SER curves for BPSK, QPSK, 8-PSK, and 16-PSK in an AWGN channel.

            .. ipython:: python

                bpsk = sdr.PSK(2); \
                qpsk = sdr.PSK(4); \
                psk8 = sdr.PSK(8); \
                psk16 = sdr.PSK(16); \
                esn0 = np.linspace(-2, 10, 100)

                @savefig sdr_psk_ser_1.png
                plt.figure(); \
                sdr.plot.ser(esn0, bpsk.ser(esn0), label="BPSK"); \
                sdr.plot.ser(esn0, qpsk.ser(esn0), label="QPSK"); \
                sdr.plot.ser(esn0, psk8.ser(esn0), label="8-PSK"); \
                sdr.plot.ser(esn0, psk16.ser(esn0), label="16-PSK"); \
                plt.title("SER curves for PSK modulation in an AWGN channel");

            Compare the symbol error rate of QPSK and DE-QPSK in an AWGN channel.

            .. ipython:: python

                @savefig sdr_psk_ser_2.png
                plt.figure(); \
                sdr.plot.ser(esn0, qpsk.ser(esn0), label="QPSK"); \
                sdr.plot.ser(esn0, qpsk.ser(esn0, diff_encoded=True), label="DE-QPSK"); \
                plt.title("SER curves for PSK and DE-PSK modulation in an AWGN channel");
        """
        esn0 = verify_arraylike(esn0, float=True)
        verify_bool(diff_encoded)

        M = self.order
        k = self.bps
        esn0_linear = linear(esn0)
        ebn0 = esn0_to_ebn0(esn0, k)
        ebn0_linear = linear(ebn0)

        if not diff_encoded:
            if M == 2:
                # Equation 4.3-13 from Proakis
                Pse = Q(np.sqrt(2 * ebn0_linear))
            elif M == 4:
                # Equation 4.3-15 from Proakis
                Pse = 2 * Q(np.sqrt(2 * ebn0_linear)) * (1 - 1 / 2 * Q(np.sqrt(2 * ebn0_linear)))
            else:
                # Equation 8.18 from Simon and Alouini
                Pse = np.zeros_like(esn0_linear)
                for i in range(esn0_linear.size):
                    Pse[i] = (
                        Q(np.sqrt(2 * esn0_linear[i]))
                        + scipy.integrate.quad(
                            lambda u, i: 2
                            / np.sqrt(np.pi)
                            * np.exp(-((u - np.sqrt(esn0_linear[i])) ** 2))
                            * Q(np.sqrt(2) * u * np.tan(np.pi / M)),
                            0,
                            np.inf,
                            args=(i,),
                        )[0]
                    )
        else:
            # Equation 8.36 from Simon and Alouini
            Pse_non_diff = self.ser(esn0, diff_encoded=False)
            Pse = 2 * Pse_non_diff - Pse_non_diff**2
            for i in range(esn0_linear.size):
                for j in range(1, M):
                    Pj = Pk(M, esn0_linear[i], j)
                    Pse[i] -= Pj**2

        return convert_output(Pse)

    @property
    @extend_docstring(
        LinearModulation.phase_offset,
        {},
        r"""
        Examples:
            See the :ref:`psk` example.

            Create a QPSK constellation with no phase offset.

            .. ipython:: python

                psk = sdr.PSK(4); \
                psk.phase_offset

                @savefig sdr_psk_phase_offset_1.png
                plt.figure(); \
                sdr.plot.symbol_map(psk.symbol_map);

            Create a QPSK constellation with 45° phase offset.

            .. ipython:: python

                psk = sdr.PSK(4, phase_offset=45); \
                psk.phase_offset

                @savefig sdr_psk_phase_offset_2.png
                plt.figure(); \
                sdr.plot.symbol_map(psk.symbol_map);
        """,
    )
    def phase_offset(self) -> float:
        return super().phase_offset

    @property
    @extend_docstring(
        LinearModulation.symbol_map,
        {},
        r"""
        Examples:
            See the :ref:`psk` example.

            The default Gray-coded symbols. Adjacent symbols only differ by one bit.

            .. ipython:: python

                psk = sdr.PSK(8); \
                psk.symbol_map

                @savefig sdr_psk_symbol_map_1.png
                plt.figure(); \
                sdr.plot.symbol_map(psk.symbol_map, annotate="bin");

            The binary-coded symbols. Adjacent symbols may differ by more than one bit.

            .. ipython:: python

                psk = sdr.PSK(8, symbol_labels="bin"); \
                psk.symbol_map

                @savefig sdr_psk_symbol_map_2.png
                plt.figure(); \
                sdr.plot.symbol_map(psk.symbol_map, annotate="bin");
        """,
    )
    def symbol_map(self) -> npt.NDArray[np.complex128]:
        return super().symbol_map


@export
class PiMPSK(PSK):
    r"""
    Implements $\pi/M$ phase-shift keying ($\pi/M$ PSK) modulation and demodulation.

    Notes:
        $\pi/M$ M-PSK is a linear phase modulation scheme similar to conventional M-PSK. One key distinction is that
        in $\pi/M$ M-PSK, the odd symbols are rotated by $\pi/M$ radians relative to the even symbols.
        This prevents symbol transitions through the origin, which results in a lower peak-to-average power ratio
        (PAPR).

        The modulation order $M = 2^k$ is a power of 2 and indicates the number of phases used $2M$.
        The input bit stream is taken $k$ bits at a time to create decimal symbols
        $s[k] \in \{0, \dots, M-1\}$. These decimal symbols $s[k]$ are then mapped to complex symbols
        $a[k] \in \mathbb{C}$ by the equation

        $$a[k] =
        \begin{cases}
        \displaystyle
        \exp \left[ j\left(\frac{2\pi}{M}s[k] + \phi\right) \right] & k\ \text{even} \\
        \displaystyle
        \exp \left[ j\left(\frac{2\pi}{M}s[k] + \phi + \pi/M\right) \right] & k\ \text{odd} \\
        \end{cases}
        $$

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Examples:
        Create a $\pi/4$ QPSK modem.

        .. ipython:: python

            pi4_qpsk = sdr.PiMPSK(4, pulse_shape="srrc"); pi4_qpsk

            @savefig sdr_PiMPSK_1.png
            plt.figure(); \
            sdr.plot.symbol_map(pi4_qpsk);

        Generate a random bit stream, convert to 2-bit symbols, and map to complex symbols.

        .. ipython:: python

            bits = np.random.randint(0, 2, 1000); bits[0:8]
            symbols = sdr.pack(bits, pi4_qpsk.bps); symbols[0:4]
            complex_symbols = pi4_qpsk.map_symbols(symbols); complex_symbols[0:4]

            @savefig sdr_PiMPSK_2.png
            plt.figure(); \
            sdr.plot.constellation(complex_symbols, linestyle="-");

        Modulate and pulse shape the symbols to a complex baseband signal.

        .. ipython:: python

            tx_samples = pi4_qpsk.modulate(symbols)

            @savefig sdr_PiMPSK_3.png
            plt.figure(); \
            sdr.plot.time_domain(tx_samples[0:50*pi4_qpsk.sps]);

        Examine the eye diagram of the pulse-shaped transmitted signal. The SRRC pulse shape is not a Nyquist filter,
        so ISI is present.

        .. ipython:: python

            @savefig sdr_PiMPSK_4.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[5*pi4_qpsk.sps : -5*pi4_qpsk.sps], pi4_qpsk.sps, persistence=True); \
            plt.suptitle("Noiseless transmitted signal with ISI");

        Add AWGN noise such that $E_b/N_0 = 30$ dB.

        .. ipython:: python

            ebn0 = 30; \
            snr = sdr.ebn0_to_snr(ebn0, bps=pi4_qpsk.bps, sps=pi4_qpsk.sps); \
            rx_samples = sdr.awgn(tx_samples, snr=snr)

            @savefig sdr_PiMPSK_5.png
            plt.figure(); \
            sdr.plot.time_domain(rx_samples[0:50*pi4_qpsk.sps]);

        Manually apply a matched filter. Examine the eye diagram of the matched filtered received signal. The
        two cascaded SRRC filters create a Nyquist RC filter. Therefore, the ISI is removed.

        .. ipython:: python

            mf = sdr.FIR(pi4_qpsk.pulse_shape); \
            mf_samples = mf(rx_samples)

            @savefig sdr_PiMPSK_6.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(mf_samples[10*pi4_qpsk.sps : -10*pi4_qpsk.sps], pi4_qpsk.sps, persistence=True); \
            plt.suptitle("Noisy received and matched filtered signal without ISI");

        Matched filter and demodulate.

        .. ipython:: python

            rx_symbols, rx_complex_symbols, _ = pi4_qpsk.demodulate(rx_samples)

            # The symbol decisions are error-free
            np.array_equal(symbols, rx_symbols)

            @savefig sdr_PiMPSK_7.png
            plt.figure(); \
            sdr.plot.constellation(rx_complex_symbols);

        See the :ref:`psk` example.

    Group:
        modulation-linear
    """

    def __init__(
        self,
        order: int,
        phase_offset: float = 0.0,
        symbol_labels: Literal["bin", "gray"] | npt.ArrayLike = "gray",
        sps: int = 8,
        pulse_shape: npt.ArrayLike | Literal["rect", "rc", "srrc"] = "rect",
        span: int | None = None,
        alpha: float | None = None,
    ):
        r"""
        Creates a new $\pi/M$ PSK object.

        Arguments:
            order: The modulation order $M = 2^k$, where $k \ge 1$ is the coded bits per symbol.
            phase_offset: The absolute phase offset $\phi$ in degrees.
            symbol_labels: The decimal symbol labels of consecutive complex symbols.

                - `"bin"`: The symbols are binary-coded. Adjacent symbols may differ by more than one bit.
                - `"gray":` The symbols are Gray-coded. Adjacent symbols only differ by one bit.
                - `npt.ArrayLike`: An $M$-length array whose indices are the default symbol labels and whose values are
                  the new symbol labels. The default symbol labels are $0$ to $M-1$ for phases starting at $1 + 0j$
                  and going counter-clockwise around the unit circle.

            sps: The number of samples per symbol $f_s / f_{sym}$.
            pulse_shape: The pulse shape $h[n]$ of the modulated signal.

                - `npt.ArrayLike`: A custom pulse shape. It is important that `sps` matches the design
                  of the pulse shape. See :ref:`pulse-shaping-functions`.
                - `"rect"`: Rectangular pulse shape.
                - `"rc"`: Raised cosine pulse shape.
                - `"srrc"`: Square-root raised cosine pulse shape.

            span: The span of the pulse shape in symbols. This is only used if `pulse_shape` is a string.
                If `None`, 1 is used for `"rect"` and 10 is used for `"rc"` and `"srrc"`.
            alpha: The roll-off factor of the pulse shape. If `None`, 0.2 is used for `"rc"` and `"srrc"`.

        See Also:
            sdr.rectangular, sdr.raised_cosine, sdr.root_raised_cosine
        """
        super().__init__(
            order,
            phase_offset=phase_offset,
            symbol_labels=symbol_labels,
            sps=sps,
            pulse_shape=pulse_shape,
        )

    def _map_symbols(self, s: npt.NDArray[np.int_]) -> npt.NDArray[np.complex128]:
        a = super()._map_symbols(s)

        # Rotate odd symbols by pi/M
        a_rotated = a.copy()
        a_rotated[1::2] *= np.exp(1j * np.pi / self.order)

        return a_rotated

    def _decide_symbols(
        self, a_tilde: npt.NDArray[np.complex128]
    ) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128]]:
        # Rotate odd symbols by -pi/M
        a_tilde_derotated = a_tilde.copy()
        a_tilde_derotated[1::2] *= np.exp(-1j * np.pi / self.order)

        return super()._decide_symbols(a_tilde_derotated)


@export
class OQPSK(PSK):
    r"""
    Implements offset quadrature phase-shift keying (OQPSK) modulation and demodulation.

    Notes:
        Offset QPSK is a linear phase modulation scheme similar to conventional QPSK. One key distinction is that
        the I and Q channels transition independently, one half symbol apart. This prevents symbol transitions
        through the origin, which results in a lower peak-to-average power ratio (PAPR).

        The modulation order $M = 2^k$ is a power of 2 and indicates the number of phases used.
        The input bit stream is taken $k$ bits at a time to create decimal symbols
        $s[k] \in \{0, \dots, M-1\}$. These decimal symbols $s[k]$ are then mapped to complex symbols
        $a[k] \in \mathbb{C}$ by the equation

        $$I[k] + jQ[k] = \exp \left[ j\left(\frac{2\pi}{M}s[k] + \phi\right) \right]$$

        $$
        \begin{align}
        a[k + 0] &= I[k] + jQ[k - 1] \\
        a[k + 1/2] &= I[k] + jQ[k] \\
        a[k + 1] &= I[k + 1] + jQ[k] \\
        a[k + 3/2] &= I[k + 1] + jQ[k + 1] \\
        \end{align}
        $$

    .. nomenclature::
        :collapsible:

        - $k$: Symbol index
        - $n$: Sample index
        - $s[k]$: Decimal symbols
        - $a[k]$ Complex symbols
        - $x[n]$: Pulse-shaped complex samples
        - $\tilde{x}[n]$: Received (noisy) pulse-shaped complex samples
        - $\tilde{a}[k]$: Received (noisy) complex symbols
        - $\hat{a}[k]$: Complex symbol decisions
        - $\hat{s}[k]$: Decimal symbol decisions

    Examples:
        Create a OQPSK modem.

        .. ipython:: python

            oqpsk = sdr.OQPSK(pulse_shape="srrc"); oqpsk

            @savefig sdr_OQPSK_1.png
            plt.figure(); \
            sdr.plot.symbol_map(oqpsk);

        Generate a random bit stream, convert to 2-bit symbols, and map to complex symbols.

        .. ipython:: python

            bits = np.random.randint(0, 2, 1000); bits[0:8]
            symbols = sdr.pack(bits, oqpsk.bps); symbols[0:4]
            complex_symbols = oqpsk.map_symbols(symbols); complex_symbols[0:4]

            @savefig sdr_OQPSK_2.png
            plt.figure(); \
            sdr.plot.constellation(complex_symbols, linestyle="-");

        Modulate and pulse shape the symbols to a complex baseband signal.

        .. ipython:: python

            tx_samples = oqpsk.modulate(symbols)

            @savefig sdr_OQPSK_3.png
            plt.figure(); \
            sdr.plot.time_domain(tx_samples[0:50*oqpsk.sps]);

        Examine the eye diagram of the pulse-shaped transmitted signal. The SRRC pulse shape is not a Nyquist filter,
        so ISI is present.

        .. ipython:: python

            @savefig sdr_OQPSK_4.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(tx_samples[5*oqpsk.sps : -5*oqpsk.sps], oqpsk.sps, persistence=True); \
            plt.suptitle("Noiseless transmitted signal with ISI");

        Add AWGN noise such that $E_b/N_0 = 30$ dB.

        .. ipython:: python

            ebn0 = 30; \
            snr = sdr.ebn0_to_snr(ebn0, bps=oqpsk.bps, sps=oqpsk.sps); \
            rx_samples = sdr.awgn(tx_samples, snr=snr)

            @savefig sdr_OQPSK_5.png
            plt.figure(); \
            sdr.plot.time_domain(rx_samples[0:50*oqpsk.sps]);

        Manually apply a matched filter. Examine the eye diagram of the matched filtered received signal. The
        two cascaded SRRC filters create a Nyquist RC filter. Therefore, the ISI is removed.

        .. ipython:: python

            mf = sdr.FIR(oqpsk.pulse_shape); \
            mf_samples = mf(rx_samples)

            @savefig sdr_OQPSK_6.png
            plt.figure(figsize=(8, 6)); \
            sdr.plot.eye(mf_samples[10*oqpsk.sps : -10*oqpsk.sps], oqpsk.sps, persistence=True); \
            plt.suptitle("Noisy received and matched filtered signal without ISI");

        Matched filter and demodulate. Note, the first symbol has $Q = 0$ and the last symbol has $I = 0$.

        .. ipython:: python

            rx_symbols, rx_complex_symbols, _ = oqpsk.demodulate(rx_samples)

            # The symbol decisions are error-free
            np.array_equal(symbols, rx_symbols)

            @savefig sdr_OQPSK_7.png
            plt.figure(); \
            sdr.plot.constellation(rx_complex_symbols);

        See the :ref:`psk` example.

    Group:
        modulation-linear
    """

    def __init__(
        self,
        phase_offset: float = 45,
        symbol_labels: Literal["bin", "gray"] | npt.ArrayLike = "gray",
        sps: int = 8,
        pulse_shape: npt.ArrayLike | Literal["rect", "rc", "srrc"] = "rect",
        span: int | None = None,
        alpha: float | None = None,
    ):
        r"""
        Creates a new OQPSK object.

        Arguments:
            phase_offset: The absolute phase offset $\phi$ in degrees.
            symbol_labels: The decimal symbol labels of consecutive complex symbols.

                - `"bin"`: The symbols are binary-coded. Adjacent symbols may differ by more than one bit.
                - `"gray":` The symbols are Gray-coded. Adjacent symbols only differ by one bit.
                - `npt.ArrayLike`: An $4$-length array whose indices are the default symbol labels and whose values are
                  the new symbol labels. The default symbol labels are $0$ to $4-1$ for phases starting at $1 + 0j$
                  and going counter-clockwise around the unit circle.

            sps: The number of samples per symbol $f_s / f_{sym}$.
            pulse_shape: The pulse shape $h[n]$ of the modulated signal.

                - `npt.ArrayLike`: A custom pulse shape. It is important that `sps` matches the design
                  of the pulse shape. See :ref:`pulse-shaping-functions`.
                - `"rect"`: Rectangular pulse shape.
                - `"rc"`: Raised cosine pulse shape.
                - `"srrc"`: Square-root raised cosine pulse shape.

            span: The span of the pulse shape in symbols. This is only used if `pulse_shape` is a string.
                If `None`, 1 is used for `"rect"` and 10 is used for `"rc"` and `"srrc"`.
            alpha: The roll-off factor of the pulse shape. If `None`, 0.2 is used for `"rc"` and `"srrc"`.

        See Also:
            sdr.rectangular, sdr.raised_cosine, sdr.root_raised_cosine
        """
        super().__init__(
            4,
            phase_offset=phase_offset,
            symbol_labels=symbol_labels,
            sps=sps,
            pulse_shape=pulse_shape,
            span=span,
            alpha=alpha,
        )

    def __repr__(self) -> str:
        return f"sdr.{type(self).__name__}(phase_offset={self.phase_offset}, symbol_labels={self._symbol_labels_str!r})"

    def __str__(self) -> str:
        string = f"sdr.{type(self).__name__}:"
        string += f"\n  symbol_map: {self.symbol_map.shape} shape"
        string += f"\n    {self.symbol_map.tolist()}"
        string += f"\n  symbol_labels: {self._symbol_labels_str!r}"
        string += f"\n  phase_offset: {self.phase_offset}"
        return string

    def _map_symbols(self, s: npt.NDArray[np.int_]) -> npt.NDArray[np.complex128]:
        a = super()._map_symbols(s)

        a_I = np.repeat(a.real, 2)
        a_Q = np.repeat(a.imag, 2)

        # Shift Q symbols by 1/2 symbol
        a_I = np.append(a_I, 0)
        a_Q = np.insert(a_Q, 0, 0)

        a = a_I + 1j * a_Q

        return a

    def _tx_pulse_shape(self, a: npt.NDArray[np.complex128]) -> npt.NDArray[np.complex128]:
        a_I, a_Q = a.real, a.imag

        # Shift Q symbols by -1/2 symbol and grab 1 sample per symbol
        a_I = a_I[:-1:2]
        a_Q = a_Q[1::2]

        x_I = self._tx_filter(a_I, mode="full")  # Complex samples
        x_Q = self._tx_filter(a_Q, mode="full")  # Complex samples

        # Shift Q symbols by 1/2 symbol
        x_I = np.append(x_I, np.zeros(self.sps // 2))
        x_Q = np.insert(x_Q, 0, np.zeros(self.sps // 2))

        x = x_I + 1j * x_Q

        return x

    def _decide_symbols(
        self, a_tilde: npt.NDArray[np.complex128]
    ) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.complex128]]:
        a_tilde = np.asarray(a_tilde)
        a_tilde_I, a_tilde_Q = a_tilde.real, a_tilde.imag

        # Shift Q symbols by -1/2 symbol and grab 1 sample per symbol
        a_tilde_I = a_tilde_I[:-1:2]
        a_tilde_Q = a_tilde_Q[1::2]

        a_tilde = a_tilde_I + 1j * a_tilde_Q

        return super()._decide_symbols(a_tilde)

    def _rx_matched_filter(self, x_tilde: npt.NDArray[np.complex128]) -> npt.NDArray[np.complex128]:
        x_tilde_I, x_tilde_Q = x_tilde.real, x_tilde.imag

        # Shift Q samples by -1/2 symbol
        x_tilde_I = x_tilde_I[: -self.sps // 2]
        x_tilde_Q = x_tilde_Q[self.sps // 2 :]

        a_tilde_I = super()._rx_matched_filter(x_tilde_I)  # Complex samples
        a_tilde_Q = super()._rx_matched_filter(x_tilde_Q)  # Complex samples

        a_tilde_I = np.repeat(a_tilde_I, 2)
        a_tilde_Q = np.repeat(a_tilde_Q, 2)

        # Shift Q symbols by 1/2 symbol
        a_tilde_I = np.append(a_tilde_I, 0)
        a_tilde_Q = np.insert(a_tilde_Q, 0, 0)

        a_tilde = a_tilde_I + 1j * a_tilde_Q

        return a_tilde


def Pk(M: int, esn0_linear: float, j: int) -> float:
    """
    Determines the probability of receiving symbol j given symbol 0 was transmitted.
    """
    # Equation 8.30 from Simon and Alouini
    A = scipy.integrate.quad(
        lambda theta, j: 1
        / (2 * np.pi)
        * np.exp(-esn0_linear * np.sin((2 * j - 1) * np.pi / M) ** 2 / np.sin(theta) ** 2),
        0,
        np.pi * (1 - (2 * j - 1) / M),
        args=(j,),
    )[0]
    B = scipy.integrate.quad(
        lambda theta, j: 1
        / (2 * np.pi)
        * np.exp(-esn0_linear * np.sin((2 * j + 1) * np.pi / M) ** 2 / np.sin(theta) ** 2),
        0,
        np.pi * (1 - (2 * j + 1) / M),
        args=(j,),
    )[0]

    # Probability of landing in decision region for symbol j when symbol 0 was transmitted
    return A - B



import numpy as np
import pytest

import sdr


def test_ppi_gains_1():
    """
    Reference:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Section 7.2.3 Example #1
    """
    BnT = 0.02  # Normalized noise bandwidth
    zeta = 1 / np.sqrt(2)  # Damping factor
    K0 = 1  # NCO gain
    Kp = 2  # PED gain
    lf = sdr.LoopFilter(BnT, zeta, K0=K0, Kp=Kp)
    K1 = lf.proportional_gain
    K2 = lf.integral_gain
    assert K1 == pytest.approx(2.6e-2, rel=3)
    assert K2 == pytest.approx(6.9e-4, rel=3)


def test_ppi_gains_2():
    """
    Reference:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Section 7.2.3 Example #2
    """
    BnT = 0.02 / 16  # Normalized noise bandwidth
    zeta = 1 / np.sqrt(2)  # Damping factor
    K0 = 16  # NCO gain
    Kp = 2  # PED gain
    lf = sdr.LoopFilter(BnT, zeta, K0=K0, Kp=Kp)
    K1 = lf.proportional_gain
    K2 = lf.integral_gain
    assert K1 == pytest.approx(1.7e-3, rel=3)
    assert K2 == pytest.approx(2.8e-6, rel=3)


def test_ppi_gains_3():
    """
    Reference:
        - Michael Rice, *Digital Communications: A Discrete-Time Approach*, Appendix C.2.1.
    """
    BnT = 0.05  # Normalized noise bandwidth
    zeta = 1  # Damping factor
    K0 = 1  # NCO gain
    Kp = 1  # PED gain
    lf = sdr.LoopFilter(BnT, zeta, K0=K0, Kp=Kp)
    K1 = lf.proportional_gain
    K2 = lf.integral_gain
    assert K1 == pytest.approx(0.1479, rel=3)
    assert K2 == pytest.approx(0.0059, rel=3)



import numpy as np
import pytest

import sdr


def test_absolute_limit():
    esn0 = -100  # dB
    C = sdr.awgn_capacity(esn0)
    ebn0 = esn0 - 10 * np.log10(C)
    ebn0_limit = 10 * np.log10(np.log(2))  # ~ -1.59 dB
    assert ebn0 == pytest.approx(ebn0_limit)


def test_vector():
    """
    These numbers were verified visually from published papers.
    """
    snr = np.linspace(-30, 20, 11)
    C = sdr.awgn_capacity(snr)
    C_truth = np.array(
        [
            1.44197417e-03,
            4.55500399e-03,
            1.43552930e-02,
            4.49155310e-02,
            1.37503524e-01,
            3.96409161e-01,
            1.00000000e00,
            2.05737321e00,
            3.45943162e00,
            5.02780767e00,
            6.65821148e00,
        ]
    )
    assert np.allclose(C, C_truth)



import numpy as np

import sdr


def test_1():
    code = sdr.binary_code(2)
    code_truth = [0, 1]
    assert isinstance(code, np.ndarray)
    assert np.array_equal(code, code_truth)


def test_2():
    code = sdr.binary_code(4)
    code_truth = [0, 1, 2, 3]
    assert isinstance(code, np.ndarray)
    assert np.array_equal(code, code_truth)


def test_3():
    code = sdr.binary_code(8)
    code_truth = [0, 1, 2, 3, 4, 5, 6, 7]
    assert isinstance(code, np.ndarray)
    assert np.array_equal(code, code_truth)


def test_4():
    code = sdr.binary_code(16)
    code_truth = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    assert isinstance(code, np.ndarray)
    assert np.array_equal(code, code_truth)



import numpy as np
import pytest

import sdr


def test_psk_rect():
    rng = np.random.default_rng(0)
    psk = sdr.PSK(2, pulse_shape="rect", sps=100)
    s = rng.integers(0, psk.order, 1_000)  # Signal is 1000 seconds long
    x = psk.modulate(s)
    t_rms = sdr.rms_integration_time(x, sample_rate=psk.sps)
    assert t_rms == pytest.approx(288.672247843467, 1e-3)


def test_psk_srrc():
    rng = np.random.default_rng(0)
    psk = sdr.PSK(2, pulse_shape="srrc", sps=100)
    s = rng.integers(0, psk.order, 1_000)  # Signal is 1000 seconds long
    x = psk.modulate(s)
    t_rms = sdr.rms_integration_time(x, sample_rate=psk.sps)
    assert t_rms == pytest.approx(288.66026000704875, 1e-3)


def test_psk_srrc_parabolic():
    rng = np.random.default_rng(0)
    psk = sdr.PSK(2, pulse_shape="srrc", sps=100)
    s = rng.integers(0, psk.order, 1_000)  # Signal is 1000 seconds long
    x = psk.modulate(s)
    y = x * np.linspace(-1, 1, len(x)) ** 2  # Parabolic pulse shape
    y *= np.sqrt(sdr.energy(x) / sdr.energy(y))  # Normalize energy
    t_rms = sdr.rms_integration_time(y, sample_rate=psk.sps)
    assert t_rms == pytest.approx(422.61669635555705, 1e-3)



"""
A module containing various helper functions for the library.
"""

from __future__ import annotations

import builtins
import inspect
import sys
from typing import Any

import numpy as np
import numpy.typing as npt

SPHINX_BUILD = hasattr(builtins, "__sphinx_build__")


def _argument_names():
    """
    Finds the source code argument names from the function that called a verification function.
    """
    frame = inspect.currentframe()
    frame = inspect.getouterframes(frame)[2]  # function() -> verify() -> _argument_name()
    string = inspect.getframeinfo(frame[0]).code_context[0].strip()
    args = string[string.find("(") + 1 : -1].split(",")
    args = [arg.strip() for arg in args]  # Strip leading/trailing whitespace
    # args = [arg.split("=")[0].strip() for arg in args]  # Remove default values and strip whitespace
    return tuple(args)


def verify_positional_args(args: tuple[Any], limit: int):
    """
    Verify a limited number of positional arguments.
    """
    if len(args) > limit:
        raise ValueError(f"A max of {limit} positional arguments are acceptable, not {_argument_names()}.")

    return len(args)


def verify_specified(arg: Any) -> Any:
    """
    Verifies that the argument is not None.
    """
    if arg is None:
        raise ValueError(f"Argument {_argument_names()[0]!r} must be provided, not {arg}.")

    return arg


def verify_not_specified(arg: Any) -> Any:
    """
    Verifies that the argument is None.
    """
    if arg is not None:
        raise ValueError(f"Argument {_argument_names()[0]!r} must not be provided, not {arg}.")

    return arg


def verify_only_one_specified(*args):
    """
    Verifies that only one of the arguments is not None.
    """
    if sum(arg is not None for arg in args) != 1:
        raise ValueError(f"Exactly one of the arguments {_argument_names()} must be provided, not {args}.")

    return args


def verify_at_least_one_specified(*args):
    """
    Verifies that at least one of the arguments is not None.
    """
    if all(arg is None for arg in args):
        raise ValueError(f"At least one of the arguments {_argument_names()} must be provided, not {args}.")

    return args


def verify_isinstance(
    arg: Any,
    types: Any,
    optional: bool = False,
) -> Any:
    """
    Verifies that the argument is an instance of the specified type(s).
    """
    if optional:
        # TODO: Can this be done in a more elegant way?
        try:
            types = list(types)
        except TypeError:
            types = [types]
        types = tuple(types + [type(None)])

    if not isinstance(arg, types):
        raise TypeError(f"Argument {_argument_names()[0]!r} must be an instance of {types}, not {type(arg)}.")

    return arg


def verify_literal(
    arg: Any,
    literals: Any,
):
    """
    Verifies that the argument is one of the specified literals.
    """
    if not arg in literals:
        raise ValueError(f"Argument {_argument_names()[0]!r} must be one of {literals}, not {arg!r}.")

    return arg


def verify_scalar(
    x: Any,
    # Data types
    optional: bool = False,
    int: bool = False,
    float: bool = False,
    complex: bool = False,
    # Value constraints
    real: bool = False,
    imaginary: bool = False,
    negative: bool = False,
    non_negative: bool = False,
    positive: bool = False,
    even: bool = False,
    odd: bool = False,
    power_of_two: bool = False,
    inclusive_min: float | None = None,
    inclusive_max: float | None = None,
    exclusive_min: float | None = None,
    exclusive_max: float | None = None,
    # Conversions
    accept_numpy: bool = True,
    convert_numpy: bool = False,
) -> Any:
    """
    Verifies that the argument is a scalar and satisfies the conditions.
    """
    if optional and x is None:
        return x

    if convert_numpy:
        x = convert_to_scalar(x)

    if int:
        if not (isinstance(x, builtins.int) or (accept_numpy and np.issubdtype(x, np.integer))):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int, not {type(x)}.")
    if float:
        if not (isinstance(x, (builtins.int, builtins.float)) or (accept_numpy and np.issubdtype(x, np.floating))):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int or float, not {type(x)}.")
    if complex:
        if not (
            isinstance(x, (builtins.int, builtins.float, builtins.complex))
            or (accept_numpy and np.issubdtype(x, np.complexfloating))
        ):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int or float or complex, not {type(x)}.")

    if real:
        if isinstance(x, builtins.complex):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be real, not complex.")
    if imaginary:
        if not isinstance(x, builtins.complex):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be complex, not real.")

    if negative:
        if x >= 0:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be negative, not {x}.")
    if non_negative:
        if x < 0:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be non-negative, not {x}.")
    if positive:
        if x <= 0:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be positive, not {x}.")
    if even:
        if x % 2 != 0:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be even, not {x}.")
    if odd:
        if x % 2 == 0:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be odd, not {x}.")
    if power_of_two:
        if not (x & (x - 1) == 0):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be a power of two, not {x}.")

    if inclusive_min is not None:
        if x < inclusive_min:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be at least {inclusive_min}, not {x}.")
    if inclusive_max is not None:
        if x > inclusive_max:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be at most {inclusive_max}, not {x}.")
    if exclusive_min is not None:
        if x <= exclusive_min:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be greater than {exclusive_min}, not {x}.")
    if exclusive_max is not None:
        if x >= exclusive_max:
            raise ValueError(f"Argument {_argument_names()[0]!r} must be less than {exclusive_max}, not {x}.")

    return x


def verify_bool(
    x: Any,
    # Conversions
    accept_numpy: bool = True,
    convert_numpy: bool = False,
) -> bool:
    """
    Verifies that the argument is a boolean.
    """
    if convert_numpy:
        x = convert_to_scalar(x)

    if not (isinstance(x, bool) or (accept_numpy and isinstance(x, np.bool_))):
        raise TypeError(f"Argument {_argument_names()[0]!r} must be a bool, not {type(x)}.")

    return x


def verify_coprime(
    x: int,
    y: int,
):
    """
    Verifies that the arguments are coprime.
    """
    if np.gcd(x, y) != 1:
        raise ValueError(
            f"Arguments {_argument_names()[0]!r} and {_argument_names()[1]!r} must be coprime, not {x} and {y}."
        )


def verify_condition(
    condition: bool,
):
    """
    Verifies that the condition is satisfied.
    """
    if not condition:
        raise ValueError(f"Arguments must satisfy the condition {_argument_names()[0]!r}.")


def verify_arraylike(
    x: npt.ArrayLike | None,
    dtype: npt.DTypeLike | None = None,
    # Data types
    optional: bool = False,
    int: bool = False,
    float: bool = False,
    complex: bool = False,
    # Value constraints
    real: bool = False,
    imaginary: bool = False,
    negative: bool = False,
    non_negative: bool = False,
    positive: bool = False,
    inclusive_min: float | None = None,
    inclusive_max: float | None = None,
    exclusive_min: float | None = None,
    exclusive_max: float | None = None,
    # Dimension and size constraints
    atleast_1d: bool = False,
    atleast_2d: bool = False,
    atleast_3d: bool = False,
    ndim: int | None = None,
    size: int | None = None,
    size_multiple: int | None = None,
    shape: tuple[int, ...] | None = None,
) -> npt.NDArray:
    """
    Converts the argument to a NumPy array and verifies the conditions.
    """
    if optional and x is None:
        return x

    x = np.asarray(x, dtype=dtype)

    if int:
        if not np.issubdtype(x.dtype, np.integer):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int, not {x.dtype}.")
    if float:
        if not (np.issubdtype(x.dtype, np.integer) or np.issubdtype(x.dtype, np.floating)):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int or float, not {x.dtype}.")
    if complex:
        if not (
            np.issubdtype(x.dtype, np.integer)
            or np.issubdtype(x.dtype, np.floating)
            or np.issubdtype(x.dtype, np.complexfloating)
        ):
            raise TypeError(f"Argument {_argument_names()[0]!r} must be an int or float or complex, not {x.dtype}.")

    if real:
        if not np.isrealobj(x):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be real, not complex.")
    if imaginary:
        if not np.iscomplexobj(x):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be complex, not real.")
    if negative:
        if np.any(x >= 0):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be negative, not {x}.")
    if non_negative:
        if np.any(x < 0):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be non-negative, not {x}.")
    if positive:
        if np.any(x <= 0):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be positive, not {x}.")

    if inclusive_min is not None:
        if np.any(x < inclusive_min):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be at least {inclusive_min}, not {x}.")
    if inclusive_max is not None:
        if np.any(x > inclusive_max):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be at most {inclusive_max}, not {x}.")
    if exclusive_min is not None:
        if np.any(x <= exclusive_min):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be greater than {exclusive_min}, not {x}.")
    if exclusive_max is not None:
        if np.any(x >= exclusive_max):
            raise ValueError(f"Argument {_argument_names()[0]!r} must be less than {exclusive_max}, not {x}.")

    if atleast_1d:
        x = np.atleast_1d(x)
    if atleast_2d:
        x = np.atleast_2d(x)
    if atleast_3d:
        x = np.atleast_3d(x)
    if ndim is not None:
        if not x.ndim == ndim:
            raise ValueError(f"Argument {_argument_names()[0]!r} must have {ndim} dimensions, not {x.ndim}.")
    if size is not None:
        if not x.size == size:
            raise ValueError(f"Argument {_argument_names()[0]!r} must have {size} elements, not {x.size}.")
    if size_multiple is not None:
        if not x.size % size_multiple == 0:
            raise ValueError(
                f"Argument {_argument_names()[0]!r} must have a size that is a multiple of {size_multiple}, not {x.size}."
            )
    if shape is not None:
        if not x.shape == shape:
            raise ValueError(f"Argument {_argument_names()[0]!r} must have shape {shape}, not {x.shape}.")

    return x


def verify_same_shape(
    x: npt.NDArray,
    y: npt.NDArray,
):
    """
    Verifies that the arguments have the same shape.
    """
    if x.shape != y.shape:
        raise ValueError(
            f"Arguments {_argument_names()[0]!r} and {_argument_names()[1]!r} must have the same shape, not {x.shape} and {y.shape}."
        )


def convert_to_scalar(x: Any):
    """
    Converts the input to a scalar if possible.
    """
    if np.isscalar(x) and hasattr(x, "item"):
        x = x.item()

    # TODO: Why is this needed? array(0) with np.int64 does not return true for np.isscalar()
    if isinstance(x, np.ndarray) and x.ndim == 0:
        x = x.item()

    return x


def convert_output(
    x: Any,
    squeeze: bool = False,
) -> Any:
    """
    Converts the output to a native Python type if scalar.
    """
    if squeeze:
        x = np.squeeze(x)

    x = convert_to_scalar(x)

    return x


def export(obj):
    """
    Marks an object for exporting into the public API.

    This decorator appends the object's name to the private module's __all__ list. The private module should
    then be imported in sdr/__init__.py using from ._private_module import *. It also modifies the object's
    __module__ to "sdr".
    """
    # Determine the private module that defined the object
    module = sys.modules[obj.__module__]

    if not SPHINX_BUILD:
        # Set the object's module to the first non-private module. This way the REPL will display the object
        # as sdr.obj and not sdr._private_module.obj.
        idx = obj.__module__.find("._")
        obj.__module__ = obj.__module__[:idx]

    # Append this object to the private module's "all" list
    public_members = getattr(module, "__all__", [])
    public_members.append(obj.__name__)
    module.__all__ = public_members

    return obj


def extend_docstring(method, replace=None, docstring=""):
    """
    A decorator to extend the docstring of `method` with the provided docstring.

    The decorator also finds and replaces and key-value pair in `replace`.
    """
    replace = {} if replace is None else replace

    def decorator(obj):
        parent_docstring = getattr(method, "__doc__", "")
        if parent_docstring is None:
            return obj
        for from_str, to_str in replace.items():
            parent_docstring = parent_docstring.replace(from_str, to_str)
        obj.__doc__ = parent_docstring + "\n" + docstring

        return obj

    return decorator



import numpy as np
import pytest

import sdr


def test_standard_limits():
    assert sdr.shannon_limit_snr(0) == -np.inf
    assert sdr.shannon_limit_snr(1) == 0
    assert sdr.shannon_limit_snr(2) == pytest.approx(4.771212547196624)


def test_vector():
    """
    These numbers were verified visually from published papers.
    """
    rate = np.array([1 / 8, 1 / 4, 1 / 3, 1 / 2, 2 / 3, 3 / 4, 7 / 8])  # Code rate, n/k
    rho = 2 * rate  # bits/2D
    ebn0 = sdr.shannon_limit_snr(rho)
    ebn0_truth = np.array([-7.23062536, -3.82775685, -2.3106528, 0.0, 1.8179847, 2.62077655, 3.73571347])
    assert np.allclose(ebn0, ebn0_truth)



"""
A module containing various scramblers.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from galois.typing import ArrayLike, PolyLike

from .._helper import convert_output, export, verify_arraylike
from .._sequence import FLFSR


@export
class AdditiveScrambler:
    r"""
    Implements an additive scrambler.

    Notes:
        .. code-block:: text
           :caption: Additive Scrambler Block Diagram

                    +--------------@<-------------@<------------@<-------------+
                    |              ^              ^             ^              |
                    |              | c_n-1        | c_n-2       | c_1          | c_0
                    |              | T[0]         | T[1]        | T[n-2]       | T[n-1]
                    |              |              |             |              |
                    |  +--------+  |  +--------+  |             |  +--------+  |
                    +->|  S[0]  |--+->|  S[1]  |--+---  ...  ---+->| S[n-1] |--+
                    |  +--------+     +--------+                   +--------+
                    |
                    v
            x[n] -->@--> y[n]

            S[k] = State vector
            T[k] = Taps vector
            x[n] = Input sequence
            y[n] = Output sequence
            @ = Finite field adder

    References:
        - https://en.wikipedia.org/wiki/Scrambler

    Examples:
        Construct the additive scrambler used in IEEE 802.11.

        .. ipython:: python

            c = galois.Poly.Degrees([7, 3, 0]); c
            scrambler = sdr.AdditiveScrambler(c)

        Scramble and descramble a sequence.

        .. ipython:: python

            x = np.random.randint(0, 2, 20); x
            y = scrambler.scramble(x); y
            xx = scrambler.descramble(y); xx
            np.array_equal(x, xx)

    Group:
        coding-scramblers
    """

    def __init__(
        self,
        characteristic_poly: PolyLike | None = None,
        feedback_poly: PolyLike | None = None,
        state: ArrayLike | None = None,
    ):
        r"""
        Creates an additive scrambler.

        Arguments:
            characteristic_poly: The characteristic polynomial
                $c(x) = x^{n} - c_{n-1} \cdot x^{n-1} - c_{n-2} \cdot x^{n-2} - \dots - c_{1} \cdot x - c_{0}$.
            feedback_poly: The feedback polynomial
                $f(x) = -c_{0} \cdot x^{n} - c_{1} \cdot x^{n-1} - \dots - c_{n-2} \cdot x^{2} - c_{n-1} \cdot x + 1$.

                .. note::
                    Either `characteristic_poly` or `feedback_poly` must be specified, but not both.

            state: The initial state vector $S = [S_0, S_1, \dots, S_{n-2}, S_{n-1}]$. The default is `None`
                which corresponds to all ones.

        See Also:
            FLFSR, galois.primitive_poly
        """
        self._lfsr = FLFSR(characteristic_poly, feedback_poly, state=state)

    def scramble(self, x: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Scrambles the binary input sequence $x[n]$.

        Arguments:
            x: The binary input sequence $x[n]$.

        Returns:
            The scrambled binary output sequence $y[n]$.
        """
        x = verify_arraylike(x, int=True)

        self.lfsr.reset()  # Set the initial state
        shift = self.lfsr.characteristic_poly.degree
        seq = self.lfsr.step(x.size + shift)
        seq = seq.view(np.ndarray)
        y = np.bitwise_xor(x, seq[shift:])

        return convert_output(y)

    def descramble(self, y: npt.ArrayLike) -> npt.NDArray[np.int_]:
        r"""
        Descrambles the binary input sequence $y[n]$.

        Arguments:
            y: The binary input sequence $y[n]$.

        Returns:
            The descrambled binary output sequence $x[n]$.
        """
        return self.scramble(y)

    @property
    def lfsr(self) -> FLFSR:
        r"""
        The Fibonacci LFSR used for scrambling.
        """
        return self._lfsr



import numpy as np

import sdr


def test_outputs():
    C = sdr.bec_capacity([0, 0.5, 1])
    assert isinstance(C, np.ndarray)
    assert np.array_equal(C, [1, 0.5, 0])

    C = sdr.bec_capacity(0.5)
    assert isinstance(C, float)
    assert C == 0.5



"""
A module containing various correlation sequences.
"""

from __future__ import annotations

import math
from typing import Any, overload

import galois
import numpy as np
import numpy.typing as npt
import scipy.linalg
from galois import Poly
from galois.typing import PolyLike
from typing_extensions import Literal

from .._data import pack, unpack
from .._helper import export, verify_bool, verify_coprime, verify_isinstance, verify_literal, verify_scalar
from ._conversion import code_to_field, code_to_sequence, sequence_to_code
from ._maximum import is_preferred_pair, m_sequence, preferred_pairs


@overload
def barker_code(length: int, output: Literal["binary"] = "binary") -> npt.NDArray[np.int_]: ...


@overload
def barker_code(length: int, output: Literal["field"]) -> galois.FieldArray: ...


@overload
def barker_code(length: int, output: Literal["bipolar"]) -> npt.NDArray[np.float64]: ...


@export
def barker_code(length: Any, output: Any = "binary") -> Any:
    r"""
    Generates the Barker code/sequence of length $n$.

    Arguments:
        length: The length $n$ of the Barker code/sequence.
        output: The output format of the Barker code/sequence.

            - `"binary"` (default): The Barker code with binary values of 0 and 1.
            - `"field"`: The Barker code as a Galois field array over $\mathrm{GF}(2)$.
            - `"bipolar"`: The Barker sequence with bipolar values of 1 and -1.

    Returns:
        The Barker code/sequence of length $n$.

    Notes:
        Barker codes only exist for length $n = 1, 2, 3, 4, 5, 7, 11, 13$.

    Examples:
        Create a Barker code and sequence of length 13.

        .. ipython:: python

            sdr.barker_code(13)
            sdr.barker_code(13, output="bipolar")
            sdr.barker_code(13, output="field")

        Barker sequences have ideally minimal auto-correlation sidelobes of +1 or -1.

        .. ipython:: python

            x = sdr.barker_code(13, output="bipolar")

            @savefig sdr_barker_code_1.png
            plt.figure(); \
            sdr.plot.correlation(x, x, mode="circular");

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True)
    verify_literal(output, ["binary", "field", "bipolar"])

    if length == 1:
        code = np.array([1])
    elif length == 2:
        code = np.array([1, 0])
    elif length == 3:
        code = np.array([1, 1, 0])
    elif length == 4:
        code = np.array([1, 1, 0, 1])
    elif length == 5:
        code = np.array([1, 1, 1, 0, 1])
    elif length == 7:
        code = np.array([1, 1, 1, 0, 0, 1, 0])
    elif length == 11:
        code = np.array([1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 0])
    elif length == 13:
        code = np.array([1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1])
    else:
        raise ValueError(f"Barker sequence of length {length} does not exist.")

    if output == "binary":
        return code
    elif output == "field":
        return code_to_field(code)
    else:
        return code_to_sequence(code)


@overload
def hadamard_code(length: int, index: int, output: Literal["binary"] = "binary") -> npt.NDArray[np.int_]: ...


@overload
def hadamard_code(length: int, index: int, output: Literal["field"]) -> galois.FieldArray: ...


@overload
def hadamard_code(length: int, index: int, output: Literal["bipolar"]) -> npt.NDArray[np.float64]: ...


@export
def hadamard_code(length: Any, index: Any, output: Any = "binary") -> Any:
    r"""
    Generates the Hadamard code/sequence of length $n = 2^m$.

    Arguments:
        length: The length $n = 2^m$ of the Hadamard code/sequence.
        index: The index $i$ in $[0, n)$ of the Hadamard code.
        output: The output format of the Hadamard code/sequence.

            - `"binary"` (default): The Hadamard code with binary values of 0 and 1.
            - `"field"`: The Hadamard code as a Galois field array over $\mathrm{GF}(2)$.
            - `"bipolar"`: The Hadamard sequence with bipolar values of 1 and -1.

    Returns:
        The Hadamard code/sequence of length $n = 2^m$ and index $i$.

    References:
        - https://www.mathworks.com/help/comm/ref/comm.hadamardcode-system-object.html

    Examples:
        Create a Hadamard code and sequence of length 16.

        .. ipython:: python

            sdr.hadamard_code(16, 4)
            sdr.hadamard_code(16, 4, output="bipolar")
            sdr.hadamard_code(16, 4, output="field")

        The Hadamard and Walsh code sets are equivalent, however they are indexed differently.

        .. ipython:: python

            np.array_equal(sdr.hadamard_code(16, 3), sdr.walsh_code(16, 8))
            np.array_equal(sdr.hadamard_code(16, 11), sdr.walsh_code(16, 9))

        Hadamard sequences have zero cross-correlation when time aligned.

        .. ipython:: python

            x1 = sdr.hadamard_code(32, 30, output="bipolar"); \
            x2 = sdr.hadamard_code(32, 18, output="bipolar"); \
            x3 = sdr.hadamard_code(32, 27, output="bipolar");

            @savefig sdr_hadamard_code_1.png
            plt.figure(); \
            sdr.plot.time_domain(x1 + 3); \
            sdr.plot.time_domain(x2 + 0); \
            sdr.plot.time_domain(x3 - 3)

        Hadamard sequence auto-correlation sidelobes are not uniform as a function of sequence index.
        In fact, the sidelobes can be quite high.

        .. ipython:: python

            @savefig sdr_hadamard_code_2.png
            plt.figure(); \
            sdr.plot.correlation(x1, x1, mode="circular"); \
            sdr.plot.correlation(x2, x2, mode="circular"); \
            sdr.plot.correlation(x3, x3, mode="circular"); \
            plt.ylim(0, 32);

        Hadamard sequences have zero cross-correlation when time aligned. However, the sidelobes can be quite
        large when time misaligned. Because of this, Hadamard sequences for spreading codes are useful only when
        precise time information is known.

        .. ipython:: python

            @savefig sdr_hadamard_code_3.png
            plt.figure(); \
            sdr.plot.correlation(x1, x2, mode="circular"); \
            sdr.plot.correlation(x1, x3, mode="circular"); \
            sdr.plot.correlation(x2, x3, mode="circular"); \
            plt.ylim(0, 32);

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True, inclusive_min=2, power_of_two=True)
    verify_scalar(index, int=True, inclusive_min=0, exclusive_max=length)
    verify_literal(output, ["binary", "field", "bipolar"])

    H = scipy.linalg.hadamard(length)
    sequence = H[index]
    sequence = sequence.astype(float)

    if output == "binary":
        return sequence_to_code(sequence)
    elif output == "field":
        return code_to_field(sequence_to_code(sequence))
    else:
        return sequence


@overload
def walsh_code(length: int, index: int, output: Literal["binary"] = "binary") -> npt.NDArray[np.int_]: ...


@overload
def walsh_code(length: int, index: int, output: Literal["field"]) -> galois.FieldArray: ...


@overload
def walsh_code(length: int, index: int, output: Literal["bipolar"]) -> npt.NDArray[np.float64]: ...


@export
def walsh_code(length: Any, index: Any, output: Any = "binary") -> Any:
    r"""
    Generates the Walsh code/sequence of length $n = 2^m$.

    Arguments:
        length: The length $n = 2^m$ of the Walsh code/sequence.
        index: The index $i$ in $[0, n)$ of the Walsh code. Indicates how many transitions there are in the code.
        output: The output format of the Walsh code/sequence.

            - `"binary"` (default): The Walsh code with binary values of 0 and 1.
            - `"field"`: The Walsh code as a Galois field array over $\mathrm{GF}(2)$.
            - `"bipolar"`: The Walsh sequence with bipolar values of 1 and -1.

    Returns:
        The Walsh code/sequence of length $n = 2^m$ and index $i$.

    References:
        - https://www.mathworks.com/help/comm/ref/comm.walshcode-system-object.html

    Examples:
        Create a Walsh code and sequence of length 16. The code index 4 indicates that there are 4 transitions
        in the code.

        .. ipython:: python

            sdr.walsh_code(16, 4)
            sdr.walsh_code(16, 4, output="bipolar")
            sdr.walsh_code(16, 4, output="field")

        The Hadamard and Walsh code sets are equivalent, however they are indexed differently.

        .. ipython:: python

            np.array_equal(sdr.hadamard_code(16, 3), sdr.walsh_code(16, 8))
            np.array_equal(sdr.hadamard_code(16, 11), sdr.walsh_code(16, 9))

        Walsh sequences have zero cross-correlation when time aligned.

        .. ipython:: python

            x1 = sdr.walsh_code(32, 10, output="bipolar"); \
            x2 = sdr.walsh_code(32, 14, output="bipolar"); \
            x3 = sdr.walsh_code(32, 18, output="bipolar");

            @savefig sdr_walsh_code_1.png
            plt.figure(); \
            sdr.plot.time_domain(x1 + 3); \
            sdr.plot.time_domain(x2 + 0); \
            sdr.plot.time_domain(x3 - 3)

        Walsh sequence auto-correlation sidelobes are not uniform as a function of sequence index.
        In fact, the sidelobes can be quite high.

        .. ipython:: python

            @savefig sdr_walsh_code_2.png
            plt.figure(); \
            sdr.plot.correlation(x1, x1, mode="circular"); \
            sdr.plot.correlation(x2, x2, mode="circular"); \
            sdr.plot.correlation(x3, x3, mode="circular"); \
            plt.ylim(0, 32);

        Walsh sequences have zero cross-correlation when time aligned. However, the sidelobes can be quite
        large when time misaligned. Because of this, Walsh sequences for spreading codes are useful only when
        precise time information is known.

        .. ipython:: python

            @savefig sdr_walsh_code_3.png
            plt.figure(); \
            sdr.plot.correlation(x1, x2, mode="circular"); \
            sdr.plot.correlation(x1, x3, mode="circular"); \
            sdr.plot.correlation(x2, x3, mode="circular"); \
            plt.ylim(0, 32);

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True, inclusive_min=2, power_of_two=True)
    verify_scalar(index, int=True, inclusive_min=0, exclusive_max=length)
    verify_literal(output, ["binary", "field", "bipolar"])

    H = scipy.linalg.hadamard(length)

    # Find the index of the Hadamard matrix that corresponds to `index` transitions
    bits = int(math.log2(length))
    index_bits = unpack(index, bits)
    h_index_bits = np.concatenate(([index_bits[0]], np.bitwise_xor(index_bits[:-1], index_bits[1:])))[::-1]
    h_index = pack(h_index_bits, bits)[0]

    sequence = H[h_index]
    sequence = sequence.astype(float)

    if output == "binary":
        return sequence_to_code(sequence)
    elif output == "field":
        return code_to_field(sequence_to_code(sequence))
    else:
        return sequence


@overload
def gold_code(
    length: int,
    index: int = 0,
    poly1: PolyLike | None = None,
    poly2: PolyLike | None = None,
    verify: bool = True,
    output: Literal["binary"] = "binary",
) -> npt.NDArray[np.int_]: ...


@overload
def gold_code(
    length: int,
    index: int = 0,
    poly1: PolyLike | None = None,
    poly2: PolyLike | None = None,
    verify: bool = True,
    output: Literal["field"] = "binary",
) -> galois.FieldArray: ...


@overload
def gold_code(
    length: int,
    index: int = 0,
    poly1: PolyLike | None = None,
    poly2: PolyLike | None = None,
    verify: bool = True,
    output: Literal["bipolar"] = "binary",
) -> npt.NDArray[np.float64]: ...


@export
def gold_code(
    length: Any, index: Any = 0, poly1: Any = None, poly2: Any = None, verify: Any = True, output: Any = "binary"
) -> Any:
    r"""
    Generates the Gold code/sequence of length $n = 2^m - 1$.

    Arguments:
        length: The length $n = 2^m - 1$ of the Gold code/sequence.
        index: The index $i$ in $[-2, n)$ of the Gold code.
        poly1: The primitive polynomial of degree $m$ over $\mathrm{GF}(2)$ for the first $m$-sequence. If `None`,
            a preferred pair is found using :func:`sdr.preferred_pairs`.
        poly2: The primitive polynomial of degree $m$ over $\mathrm{GF}(2)$ for the second $m$-sequence. If `None`,
            a preferred pair is found using :func:`sdr.preferred_pairs`.
        verify: Indicates whether to verify that the provided polynomials are a preferred pair using
            :func:`sdr.is_preferred_pair`.
        output: The output format of the Gold code/sequence.

            - `"binary"` (default): The Gold code with binary values of 0 and 1.
            - `"field"`: The Gold code as a Galois field array over $\mathrm{GF}(2)$.
            - `"bipolar"`: The Gold sequence with bipolar values of 1 and -1.

    Returns:
        The Gold code/sequence of length $n = 2^m - 1$ and index $i$.

    See Also:
        sdr.preferred_pairs, sdr.is_preferred_pair

    Notes:
        Gold codes are generated by combining two preferred pair $m$-sequences, $u$ and $v$, using the formula

        $$
        c = \begin{cases}
        u & \text{if $i = -2$} \\
        v & \text{if $i = -1$} \\
        u \oplus T^i v & \text{otherwise} ,
        \end{cases}
        $$

        where $i$ is the code index, $\oplus$ is addition in $\mathrm{GF}(2)$ (or the XOR operation), and $T^i$ is a
        left shift by $i$ positions.

        Gold codes are PN sequence with good auto-correlation and cross-correlation properties. The Gold code set
        contains $2^m + 1$ sequences of length $2^m - 1$. The correlation sides are guaranteed to be less than
        or equal to $t(m)$.

        $$
        t(m) = \begin{cases}
        2^{(m+1)/2} + 1 & \text{if $m$ is odd} \\
        2^{(m+2)/2} + 1 & \text{if $m$ is even}
        \end{cases}
        $$

        There are no preferred pairs with degree $m$ divisible by 4. Therefore, there are no Gold codes with degree
        $m$ divisible by 4.

    References:
        - John Proakis, *Digital Communications*, Chapter 12.2-5: Generation of PN Sequences.

    Examples:
        Create a Gold code and sequence of length 7.

        .. ipython:: python

            sdr.gold_code(7, 1)
            sdr.gold_code(7, 1, output="bipolar")
            sdr.gold_code(7, 1, output="field")

        Create several Gold codes of length 63.

        .. ipython:: python

            x1 = sdr.gold_code(63, 0, output="bipolar"); \
            x2 = sdr.gold_code(63, 1, output="bipolar"); \
            x3 = sdr.gold_code(63, 2, output="bipolar");

            @savefig sdr_gold_code_1.png
            plt.figure(); \
            sdr.plot.time_domain(x1 + 3); \
            sdr.plot.time_domain(x2 + 0); \
            sdr.plot.time_domain(x3 - 3)

        Examine the auto-correlation of the Gold sequences.

        .. ipython:: python

            @savefig sdr_gold_code_2.png
            plt.figure(); \
            sdr.plot.correlation(x1, x1, mode="circular"); \
            sdr.plot.correlation(x2, x2, mode="circular"); \
            sdr.plot.correlation(x3, x3, mode="circular"); \
            plt.ylim(0, 63);

        Examine the cross-correlation of the Gold sequences.

        .. ipython:: python

            @savefig sdr_gold_code_3.png
            plt.figure(); \
            sdr.plot.correlation(x1, x2, mode="circular"); \
            sdr.plot.correlation(x1, x3, mode="circular"); \
            sdr.plot.correlation(x2, x3, mode="circular"); \
            plt.ylim(0, 63);

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True, positive=True)
    verify_scalar(length + 1, power_of_two=True)
    m = int(np.log2(length + 1))
    verify_scalar(index, int=True, inclusive_min=-2, exclusive_max=length)
    verify_bool(verify)
    verify_literal(output, ["binary", "field", "bipolar"])

    if poly1 is None and poly2 is None:
        poly1, poly2 = next(preferred_pairs(m))
    elif poly1 is None:
        poly1 = next(preferred_pairs(m, poly2))[1]
    elif poly2 is None:
        poly2 = next(preferred_pairs(m, poly1))[1]
    else:
        if verify and not is_preferred_pair(poly1, poly2):
            raise ValueError(
                f"Arguments 'poly1' and 'poly2' must be a preferred pair to generate a Gold code, {poly1} and {poly2} are not."
                + " You can pass 'verify=False' to disable this check."
            )

    u = m_sequence(m, poly=poly1, index=1, output="decimal")
    v = m_sequence(m, poly=poly2, index=1, output="decimal")

    if index == -2:
        code = u
    elif index == -1:
        code = v
    else:
        code = np.bitwise_xor(u, np.roll(v, -index))

    if output == "binary":
        return code
    elif output == "field":
        return code_to_field(code)
    else:
        return code_to_sequence(code)


@overload
def kasami_code(
    length: int, index: int | tuple[int, int] = 0, poly: PolyLike | None = None, output: Literal["binary"] = "binary"
) -> npt.NDArray[np.int_]: ...


@overload
def kasami_code(
    length: int, index: int | tuple[int, int] = 0, poly: PolyLike | None = None, output: Literal["field"] = "binary"
) -> galois.FieldArray: ...


@overload
def kasami_code(
    length: int, index: int | tuple[int, int] = 0, poly: PolyLike | None = None, output: Literal["bipolar"] = "binary"
) -> npt.NDArray[np.float64]: ...


@export
def kasami_code(length: Any, index: Any = 0, poly: Any = None, output: Any = "binary") -> Any:
    r"""
    Generates the Kasami code/sequence of length $n = 2^m - 1$.

    Arguments:
        length: The length $n = 2^m - 1$ of the Kasami code/sequence. The degree $m$ must be even.
        index: The index of the Kasami code.

            - `int`: The index $j$ in $[-1, 2^{m/2} - 1)$ from the Kasami code small set. There are $2^{m/2}$ codes
              in the small set.
            - `tuple[int, int]`: The index $(i, j)$ from the Kasami code large set, with $i \in [-2, 2^m - 1)$ and
              $j \in [-1, 2^{m/2} - 1)$. There are $(2^m + 1) \cdot 2^{m/2}$ codes in the large set.

        poly: The primitive polynomial of degree $m$ over $\mathrm{GF}(2)$. The default is `None`, which uses the
            default primitive polynomial of degree $m$, i.e. `galois.primitive_poly(2, m)`.

        output: The output format of the Kasami code/sequence.

            - `"binary"` (default): The Kasami code with binary values of 0 and 1.
            - `"field"`: The Kasami code as a Galois field array over $\mathrm{GF}(2)$.
            - `"bipolar"`: The Kasami sequence with bipolar values of 1 and -1.

    Returns:
        The Kasami code/sequence of length $n = 2^m - 1$.

    References:
        - https://en.wikipedia.org/wiki/Kasami_code

    Examples:
        Create a Kasami code and sequence of length 15.

        .. ipython:: python

            sdr.kasami_code(15, 1)
            sdr.kasami_code(15, 1, output="bipolar")
            sdr.kasami_code(15, 1, output="field")

        Create several Kasami codes of length 63 from the small set.

        .. ipython:: python

            x1 = sdr.kasami_code(63, 0, output="bipolar"); \
            x2 = sdr.kasami_code(63, 1, output="bipolar"); \
            x3 = sdr.kasami_code(63, 2, output="bipolar");

            @savefig sdr_kasami_code_1.png
            plt.figure(); \
            sdr.plot.time_domain(x1 + 3); \
            sdr.plot.time_domain(x2 + 0); \
            sdr.plot.time_domain(x3 - 3)

        Examine the auto-correlation of the Kasami sequences.

        .. ipython:: python

            @savefig sdr_kasami_code_2.png
            plt.figure(); \
            sdr.plot.correlation(x1, x1, mode="circular"); \
            sdr.plot.correlation(x2, x2, mode="circular"); \
            sdr.plot.correlation(x3, x3, mode="circular"); \
            plt.ylim(0, 63);

        Examine the cross-correlation of the Kasami sequences.

        .. ipython:: python

            @savefig sdr_kasami_code_3.png
            plt.figure(); \
            sdr.plot.correlation(x1, x2, mode="circular"); \
            sdr.plot.correlation(x1, x3, mode="circular"); \
            sdr.plot.correlation(x2, x3, mode="circular"); \
            plt.ylim(0, 63);

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True, positive=True)
    verify_scalar(length + 1, power_of_two=True)
    n = int(np.log2(length + 1))
    verify_scalar(n, even=True)
    verify_isinstance(index, (int, tuple))

    if poly is None:
        c = galois.primitive_poly(2, n)
    else:
        c = Poly._PolyLike(poly, field=galois.GF(2))

    if isinstance(index, int):
        code = _kasami_small_set(n, c, index)
    elif isinstance(index, tuple):
        code = _kasami_large_set(n, c, index)

    if output == "binary":
        return code
    elif output == "field":
        return code_to_field(code)
    else:
        return code_to_sequence(code)


def _kasami_small_set(degree: int, poly: Poly, index: int) -> npt.NDArray[np.int_]:
    if not degree % 2 == 0:
        raise ValueError(f"Argument 'degree' must be even, not {degree}.")

    j = index
    if not -1 <= j < 2 ** (degree // 2) - 1:
        raise ValueError(f"Argument 'index' must be between -1 and {2**(degree//2) - 1}, not {j}.")

    u = m_sequence(degree, poly=poly, index=1, output="decimal")
    length = u.size

    stride = 2 ** (degree // 2) + 1
    idxs = (np.arange(0, length) * stride) % length
    w = u[idxs]

    if j == -1:
        code = u
    else:
        code = np.bitwise_xor(u, np.roll(w, -j))

    return code


def _kasami_large_set(degree: int, poly: Poly, index: tuple[int, int]) -> npt.NDArray[np.int_]:
    if not degree % 4 == 2:
        raise ValueError(f"Argument 'degree' must be 2 mod 4, not {degree}.")

    i, j = index
    if not -2 <= i < 2**degree - 1:
        raise ValueError(f"Argument 'index[0]' must be between -2 and {2**degree - 1}, not {i}.")
    if not -1 <= j < 2 ** (degree // 2) - 1:
        raise ValueError(f"Argument 'index[1]' must be between -1 and {2**(degree//2) - 1}, not {j}.")

    u = m_sequence(degree, poly=poly, index=1, output="decimal")
    length = u.size

    stride = 2 ** (degree // 2) + 1
    idxs = (np.arange(0, length) * stride) % length
    w = u[idxs]

    stride = 2 ** (degree // 2 + 1) + 1
    idxs = (np.arange(0, length) * stride) % length
    v = u[idxs]

    if j == -1:
        if i == -2:
            code = u
        elif i == -1:
            code = v
        else:
            code = np.bitwise_xor(u, np.roll(v, -i))
    else:
        if i == -2:
            code = np.bitwise_xor(u, np.roll(w, -j))
        elif i == -1:
            code = np.bitwise_xor(v, np.roll(w, -j))
        else:
            code = np.bitwise_xor(np.bitwise_xor(u, np.roll(v, -i)), np.roll(w, -j))

    return code


@export
def zadoff_chu_sequence(length: int, root: int, shift: int = 0) -> npt.NDArray[np.complex128]:
    r"""
    Generates the root-$u$ Zadoff-Chu sequence of length $N$.

    Arguments:
        length: The length $N$ of the Zadoff-Chu sequence.
        root: The root $0 < u < N$ of the Zadoff-Chu sequence. The root must be relatively prime to the length,
            i.e., $\gcd(u, N) = 1$.
        shift: The shift $q \in \mathbb{Z}$ of the Zadoff-Chu sequence. When $q \ne 0$, the returned sequence
            is a cyclic shift of the root-$u$ Zadoff-Chu sequence.

    Returns:
        The root-$u$ Zadoff-Chu sequence of length $N$.

    Notes:
        The root-$u$ Zadoff-Chu sequence with length $N$ and shift $q$ is defined as

        $$x_u[n] = \exp \left( -j \frac{\pi u n (n + c_{f} + 2q)}{N} \right) ,$$

        where $c_{f} = N \mod 2$.

    References:
        - https://en.wikipedia.org/wiki/Zadoff%E2%80%93Chu_sequence

    Examples:
        Create a root-3 Zadoff-Chu sequence $x_3[n]$ with length 139.

        .. ipython:: python

            N = 139
            x3 = sdr.zadoff_chu_sequence(N, 3)

            @savefig sdr_zadoff_chu_1.png
            plt.figure(); \
            sdr.plot.constellation(x3, linestyle="-", linewidth=0.5); \
            plt.title(f"Root-3 Zadoff-Chu sequence of length {N}");

        The *periodic* auto-correlation of a Zadoff-Chu sequence has sidelobes with magnitude 0.

        .. ipython:: python

            @savefig sdr_zadoff_chu_2.png
            plt.figure(); \
            sdr.plot.correlation(x3, x3, mode="circular"); \
            plt.ylim(0, N);

        Create a root-5 Zadoff-Chu sequence $x_5[n]$ with length 139.

        .. ipython:: python

            x5 = sdr.zadoff_chu_sequence(N, 5)

            @savefig sdr_zadoff_chu_3.png
            plt.figure(); \
            sdr.plot.constellation(x5, linestyle="-", linewidth=0.5); \
            plt.title(f"Root-5 Zadoff-Chu sequence of length {N}");

        The *periodic* cross-correlation of two prime-length Zadoff-Chu sequences with different roots has sidelobes
        with magnitude $1 / \sqrt{N}$.

        .. ipython:: python

            @savefig sdr_zadoff_chu_4.png
            plt.figure(); \
            sdr.plot.correlation(x3, x5, mode="circular"); \
            plt.ylim(0, N);

    Group:
        sequences-correlation
    """
    verify_scalar(length, int=True, inclusive_min=2)
    verify_scalar(root, int=True, exclusive_min=0, exclusive_max=length)
    verify_coprime(length, root)
    verify_scalar(shift, int=True)

    n = np.arange(length)
    cf = length % 2
    xu = np.exp(-1j * np.pi * root * n * (n + cf + 2 * shift) / length)

    return xu



"""
A module containing conversions between binary and bipolar sequences.
"""

from __future__ import annotations

import galois
import numpy as np
import numpy.typing as npt


def code_to_sequence(code: npt.NDArray[np.int_]) -> npt.NDArray[np.float64]:
    r"""
    Converts a binary code to a bipolar sequence.

    Notes:
        The mapping is equivalent to BPSK: 0 -> 1, 1 -> -1.
    """
    sequence = 1 - 2 * code
    sequence = sequence.astype(float)
    return sequence


def code_to_field(code: npt.NDArray[np.int_]) -> galois.FieldArray:
    r"""
    Converts a binary code to a Galois field array over $\mathrm{GF}(2)$.
    """
    GF = galois.GF(2)
    field = GF(code)
    return field


def sequence_to_code(sequence: npt.NDArray[np.float64]) -> npt.NDArray[np.int_]:
    r"""
    Converts a bipolar sequence to a binary code.

    Notes:
        The mapping is equivalent to BPSK: 1 -> 0, -1 -> 1.
    """
    code = (1 - sequence) / 2
    code = code.astype(int)
    return code


def field_to_code(field: galois.FieldArray) -> npt.NDArray[np.int_]:
    r"""
    Converts a Galois field array over $\mathrm{GF}(2)$ to a binary code.
    """
    code = field.view(np.ndarray)
    code = code.astype(int)
    return code



import numpy as np
import pytest

import sdr


def test_scalar():
    assert sdr.coherent_gain_loss(1e-3, 235) == pytest.approx(0.8038919141626675)


def test_time_vector():
    cgl = sdr.coherent_gain_loss([1e-3, 2e-3, 3e-3, 4e-4, 5e-3], 100)
    cgl_truth = np.array([0.14335017, 0.57922366, 1.32626966, 0.02287239, 3.92239754])
    assert isinstance(cgl, np.ndarray)
    assert np.allclose(cgl, cgl_truth)


def test_freq_vector():
    cgl = sdr.coherent_gain_loss(1e-3, [0, 100, 200, 300, 400, 500])
    cgl_truth = np.array([0.0, 0.14335017, 0.57922366, 1.32626966, 2.42007077, 3.92239754])
    assert isinstance(cgl, np.ndarray)
    assert np.allclose(cgl, cgl_truth)



"""
A module for designing finite impulse response (FIR) filters.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy.signal
from typing_extensions import Literal

from .._helper import export, verify_scalar


def _normalize(h: npt.NDArray[np.float64], norm: Literal["power", "energy", "passband"]) -> npt.NDArray[np.float64]:
    if norm == "power":
        h /= np.sqrt(np.max(np.abs(h) ** 2))
    elif norm == "energy":
        h /= np.sqrt(np.sum(np.abs(h) ** 2))
    elif norm == "passband":
        h /= np.sum(h)
    else:
        raise ValueError(f"Argument 'norm' must be 'power', 'energy', or 'passband', not {norm}.")

    return h


def _normalize_passband(h: npt.NDArray[np.float64], center_freq: float) -> npt.NDArray[np.float64]:
    order = h.size - 1
    lo = np.exp(-1j * np.pi * center_freq * np.arange(-order // 2, order // 2 + 1))
    h = h / np.abs(np.sum(h * lo))
    return h


def _ideal_lowpass(order: int, cutoff_freq: float) -> npt.NDArray[np.float64]:
    """
    Returns the ideal lowpass filter impulse response.
    """
    wc = np.pi * cutoff_freq  # Cutoff frequency in radians/s
    n = np.arange(-order // 2, order // 2 + 1)  # Sample indices
    h_ideal = wc / np.pi * np.sinc(wc * n / np.pi)  # Ideal filter impulse response
    return h_ideal


def _ideal_highpass(order: int, cutoff_freq: float) -> npt.NDArray[np.float64]:
    """
    Returns the ideal highpass filter impulse response.
    """
    lo = np.cos(np.pi * np.arange(-order // 2, order // 2 + 1))
    h_ideal = _ideal_lowpass(order, 1 - cutoff_freq) * lo
    return h_ideal


def _ideal_bandpass(order: int, center_freq: float, bandwidth: float) -> npt.NDArray[np.float64]:
    """
    Returns the ideal bandpass filter impulse response.
    """
    h_lp1 = _ideal_lowpass(order, center_freq + bandwidth / 2)
    h_lp2 = _ideal_lowpass(order, center_freq - bandwidth / 2)
    h_ideal = h_lp1 - h_lp2
    return h_ideal


def _ideal_bandstop(order: int, center_freq: float, bandwidth: float) -> npt.NDArray[np.float64]:
    """
    Returns the ideal bandstop filter impulse response.
    """
    h_lp = _ideal_lowpass(order, center_freq - bandwidth / 2)
    h_hp = _ideal_highpass(order, center_freq + bandwidth / 2)
    h_ideal = h_lp + h_hp
    return h_ideal


@export
def lowpass_fir(
    order: int,
    cutoff_freq: float,
    window: str | float | tuple | None = "hamming",
) -> npt.NDArray[np.float64]:
    r"""
    Designs a lowpass FIR filter impulse response $h[n]$ using the window method.

    Arguments:
        order: The filter order $N$. Must be even.
        cutoff_freq: The cutoff frequency $f_c$, normalized to the Nyquist frequency $f_s / 2$.
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.

    Returns:
        The filter impulse response $h[n]$ with length $N + 1$. The center of the passband has 0 dB gain.

    References:
        - https://www.mathworks.com/help/dsp/ref/designlowpassfir.html

    Examples:
        Design a length-101 lowpass FIR filter with cutoff frequency $f_c = 0.2 \cdot f_s / 2$, using a Hamming window.

        .. ipython:: python

            h_hamming = sdr.lowpass_fir(100, 0.2, window="hamming")

            @savefig sdr_lowpass_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_hamming);

            @savefig sdr_lowpass_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming);

        Compare filter designs using different windows.

        .. ipython:: python

            h_hann = sdr.lowpass_fir(100, 0.2, window="hann"); \
            h_blackman = sdr.lowpass_fir(100, 0.2, window="blackman"); \
            h_blackman_harris = sdr.lowpass_fir(100, 0.2, window="blackmanharris"); \
            h_chebyshev = sdr.lowpass_fir(100, 0.2, window=("chebwin", 60)); \
            h_kaiser = sdr.lowpass_fir(100, 0.2, window=("kaiser", 0.5))

            @savefig sdr_lowpass_fir_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming, label="Hamming"); \
            sdr.plot.magnitude_response(h_hann, label="Hann"); \
            sdr.plot.magnitude_response(h_blackman, label="Blackman"); \
            sdr.plot.magnitude_response(h_blackman_harris, label="Blackman-Harris"); \
            sdr.plot.magnitude_response(h_chebyshev, label="Chebyshev"); \
            sdr.plot.magnitude_response(h_kaiser, label="Kaiser"); \
            plt.ylim(-100, 10);

    Group:
        dsp-fir-filtering
    """
    verify_scalar(order, int=True, even=True)
    verify_scalar(cutoff_freq, float=True, inclusive_min=0, inclusive_max=1)

    h = _ideal_lowpass(order, cutoff_freq)
    if window is not None:
        h *= scipy.signal.windows.get_window(window, h.size, fftbins=False)
    h = _normalize_passband(h, 0)

    return h


@export
def highpass_fir(
    order: int,
    cutoff_freq: float,
    window: str | float | tuple | None = "hamming",
) -> npt.NDArray[np.float64]:
    r"""
    Designs a highpass FIR filter impulse response $h[n]$ using the window method.

    Arguments:
        order: The filter order $N$. Must be even.
        cutoff_freq: The cutoff frequency $f_c$, normalized to the Nyquist frequency $f_s / 2$.
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.

    Returns:
        The filter impulse response $h[n]$ with length $N + 1$. The center of the passband has 0 dB gain.

    References:
        - https://www.mathworks.com/help/dsp/ref/designhighpassfir.html

    Examples:
        Design a length-101 highpass FIR filter with cutoff frequency $f_c = 0.7 \cdot f_s / 2$, using a Hamming window.

        .. ipython:: python

            h_hamming = sdr.highpass_fir(100, 0.7, window="hamming")

            @savefig sdr_highpass_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_hamming);

            @savefig sdr_highpass_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming);

        Compare filter designs using different windows.

        .. ipython:: python

            h_hann = sdr.highpass_fir(100, 0.7, window="hann"); \
            h_blackman = sdr.highpass_fir(100, 0.7, window="blackman"); \
            h_blackman_harris = sdr.highpass_fir(100, 0.7, window="blackmanharris"); \
            h_chebyshev = sdr.highpass_fir(100, 0.7, window=("chebwin", 60)); \
            h_kaiser = sdr.highpass_fir(100, 0.7, window=("kaiser", 0.5))

            @savefig sdr_highpass_fir_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming, label="Hamming"); \
            sdr.plot.magnitude_response(h_hann, label="Hann"); \
            sdr.plot.magnitude_response(h_blackman, label="Blackman"); \
            sdr.plot.magnitude_response(h_blackman_harris, label="Blackman-Harris"); \
            sdr.plot.magnitude_response(h_chebyshev, label="Chebyshev"); \
            sdr.plot.magnitude_response(h_kaiser, label="Kaiser"); \
            plt.ylim(-100, 10);

    Group:
        dsp-fir-filtering
    """
    verify_scalar(order, int=True, even=True)
    verify_scalar(cutoff_freq, float=True, inclusive_min=0, inclusive_max=1)

    h = _ideal_highpass(order, cutoff_freq)
    if window is not None:
        h *= scipy.signal.windows.get_window(window, h.size, fftbins=False)
    h = _normalize_passband(h, 1)

    return h


@export
def bandpass_fir(
    order: int,
    center_freq: float,
    bandwidth: float,
    window: str | float | tuple | None = "hamming",
) -> npt.NDArray[np.float64]:
    r"""
    Designs a bandpass FIR filter impulse response $h[n]$ using the window method.

    Arguments:
        order: The filter order $N$. Must be even.
        center_freq: The center frequency $f_{center}$, normalized to the Nyquist frequency $f_s / 2$.
        bandwidth: The two-sided bandwidth about $f_{center}$, normalized to the Nyquist frequency $f_s / 2$.
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.

    Returns:
        The filter impulse response $h[n]$ with length $N + 1$. The center of the passband has 0 dB gain.

    References:
        - https://www.mathworks.com/help/dsp/ref/designbandpassfir.html

    Examples:
        Design a length-101 bandpass FIR filter with center frequency $f_{center} = 0.4 \cdot f_s / 2$
        and bandwidth $0.1 \cdot f_s / 2$, using a Hamming window.

        .. ipython:: python

            h_hamming = sdr.bandpass_fir(100, 0.4, 0.1, window="hamming")

            @savefig sdr_bandpass_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_hamming);

            @savefig sdr_bandpass_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming);

        Compare filter designs using different windows.

        .. ipython:: python

            h_hann = sdr.bandpass_fir(100, 0.4, 0.1, window="hann"); \
            h_blackman = sdr.bandpass_fir(100, 0.4, 0.1, window="blackman"); \
            h_blackman_harris = sdr.bandpass_fir(100, 0.4, 0.1, window="blackmanharris"); \
            h_chebyshev = sdr.bandpass_fir(100, 0.4, 0.1, window=("chebwin", 60)); \
            h_kaiser = sdr.bandpass_fir(100, 0.4, 0.1, window=("kaiser", 0.5))

            @savefig sdr_bandpass_fir_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming, label="Hamming"); \
            sdr.plot.magnitude_response(h_hann, label="Hann"); \
            sdr.plot.magnitude_response(h_blackman, label="Blackman"); \
            sdr.plot.magnitude_response(h_blackman_harris, label="Blackman-Harris"); \
            sdr.plot.magnitude_response(h_chebyshev, label="Chebyshev"); \
            sdr.plot.magnitude_response(h_kaiser, label="Kaiser"); \
            plt.ylim(-100, 10);

    Group:
        dsp-fir-filtering
    """
    verify_scalar(order, int=True, even=True)
    verify_scalar(center_freq, float=True, inclusive_min=0, inclusive_max=1)
    verify_scalar(bandwidth, float=True, inclusive_min=0, inclusive_max=2 * min(center_freq, 1 - center_freq))

    h = _ideal_bandpass(order, center_freq, bandwidth)
    if window is not None:
        h *= scipy.signal.windows.get_window(window, h.size, fftbins=False)
    h = _normalize_passband(h, center_freq)

    return h


@export
def bandstop_fir(
    order: int,
    center_freq: float,
    bandwidth: float,
    window: str | float | tuple | None = "hamming",
) -> npt.NDArray[np.float64]:
    r"""
    Designs a bandstop FIR filter impulse response $h[n]$ using the window method.

    Arguments:
        order: The filter order $N$. Must be even.
        center_freq: The center frequency $f_{center}$, normalized to the Nyquist frequency $f_s / 2$.
        bandwidth: The two-sided bandwidth about $f_{center}$, normalized to the Nyquist frequency $f_s / 2$.
        window: The SciPy window definition. See :func:`scipy.signal.windows.get_window` for details.
            If `None`, no window is applied.

    Returns:
        The filter impulse response $h[n]$ with length $N + 1$. The center of the larger passband has 0 dB gain.

    References:
        - https://www.mathworks.com/help/dsp/ref/designbandstopfir.html

    Examples:
        Design a length-101 bandstop FIR filter with center frequency $f_{center} = 0.4 \cdot f_s / 2$
        and bandwidth $0.75 \cdot f_s / 2$, using a Hamming window.

        .. ipython:: python

            h_hamming = sdr.bandstop_fir(100, 0.4, 0.75, window="hamming")

            @savefig sdr_bandstop_fir_1.png
            plt.figure(); \
            sdr.plot.impulse_response(h_hamming);

            @savefig sdr_bandstop_fir_2.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming);

        Compare filter designs using different windows.

        .. ipython:: python

            h_hann = sdr.bandstop_fir(100, 0.4, 0.75, window="hann"); \
            h_blackman = sdr.bandstop_fir(100, 0.4, 0.75, window="blackman"); \
            h_blackman_harris = sdr.bandstop_fir(100, 0.4, 0.75, window="blackmanharris"); \
            h_chebyshev = sdr.bandstop_fir(100, 0.4, 0.75, window=("chebwin", 60)); \
            h_kaiser = sdr.bandstop_fir(100, 0.4, 0.75, window=("kaiser", 0.5))

            @savefig sdr_bandstop_fir_3.png
            plt.figure(); \
            sdr.plot.magnitude_response(h_hamming, label="Hamming"); \
            sdr.plot.magnitude_response(h_hann, label="Hann"); \
            sdr.plot.magnitude_response(h_blackman, label="Blackman"); \
            sdr.plot.magnitude_response(h_blackman_harris, label="Blackman-Harris"); \
            sdr.plot.magnitude_response(h_chebyshev, label="Chebyshev"); \
            sdr.plot.magnitude_response(h_kaiser, label="Kaiser"); \
            plt.ylim(-100, 10);

    Group:
        dsp-fir-filtering
    """
    verify_scalar(order, int=True, even=True)
    verify_scalar(center_freq, float=True, inclusive_min=0, inclusive_max=1)
    verify_scalar(bandwidth, float=True, inclusive_min=0, inclusive_max=2 * min(center_freq, 1 - center_freq))

    h = _ideal_bandstop(order, center_freq, bandwidth)
    if window is not None:
        h *= scipy.signal.windows.get_window(window, h.size, fftbins=False)
    if center_freq > 0.5:
        h = _normalize_passband(h, 0)
    else:
        h = _normalize_passband(h, 1)

    return h
